from __future__ import annotations

import pytest

from daft.expressions import col
from daft.functions import (
    columns_max,
    columns_mean,
    columns_min,
    columns_sum,
)
from daft.functions.functions import columns_avg


def test_columns_sum(make_df) -> None:
    data = {"a": [1, 2, 3], "b": [4, 5, 6], "c": [7, 8, 9]}
    df = make_df(data).with_column("sum", columns_sum("a", "b", "c"))

    assert df.to_pydict()["sum"] == [12, 15, 18]


def test_columns_sum_with_nulls(make_df) -> None:
    data = {"a": [1, None, 3], "b": [4, 5, None], "c": [None, 8, 9]}
    df = make_df(data).with_column("sum", columns_sum("a", "b", "c"))

    assert df.to_pydict()["sum"] == [5, 13, 12]


def test_columns_sum_with_expressions(make_df) -> None:
    data = {"a": [1, 2, 3], "b": [1, 2, 3], "c": [1, 2, 3]}
    df = make_df(data).with_column("sum", columns_sum(col("a") + 1, col("b") + 2, col("c") + 3))

    assert df.to_pydict()["sum"] == [9, 12, 15]


def test_columns_sum_in_select(make_df) -> None:
    df = make_df({"a": [1, 2, 3], "b": [4, 5, 6], "c": [7, 8, 9]})
    df = df.select(columns_sum("a", "b", "c"))

    assert df.to_pydict()["columns_sum"] == [12, 15, 18]


def test_columns_sum_with_no_input_columns(make_df) -> None:
    df = make_df({"a": [1, 2, 3]})

    with pytest.raises(ValueError, match="columns_sum requires at least one expression"):
        df.with_column("sum", columns_sum()).collect()


def test_columns_mean(make_df) -> None:
    data = {"a": [1, 2, 3], "b": [4, 5, 6], "c": [7, 8, 9]}
    df = make_df(data).with_columns(
        {
            "mean": columns_mean("a", "b", "c"),
            "avg": columns_avg("a", "b", "c"),
        }
    )

    assert df.to_pydict()["mean"] == [4.0, 5.0, 6.0]
    assert df.to_pydict()["avg"] == [4.0, 5.0, 6.0]


def test_columns_mean_with_nulls(make_df) -> None:
    data = {"a": [1, None, None], "b": [None, 5, None], "c": [None, None, 9]}
    df = make_df(data).with_columns(
        {
            "mean": columns_mean("a", "b", "c"),
            "avg": columns_avg("a", "b", "c"),
        }
    )

    assert df.to_pydict()["mean"] == [1.0, 5.0, 9.0]
    assert df.to_pydict()["avg"] == [1.0, 5.0, 9.0]


def test_columns_mean_with_expressions(make_df) -> None:
    data = {"a": [1, 2, 3], "b": [1, 2, 3], "c": [1, 2, 3]}
    df = make_df(data).with_columns(
        {
            "mean": columns_mean(col("a") + 1, col("b") + 2, col("c") + 3),
            "avg": columns_avg(col("a") + 1, col("b") + 2, col("c") + 3),
        }
    )

    assert df.to_pydict()["mean"] == [3.0, 4.0, 5.0]
    assert df.to_pydict()["avg"] == [3.0, 4.0, 5.0]


def test_columns_mean_in_select(make_df) -> None:
    df = make_df({"a": [1, 2, 3], "b": [4, 5, 6], "c": [7, 8, 9]})
    df = df.select(
        columns_mean("a", "b", "c"),
        columns_avg("a", "b", "c"),
    )

    assert df.to_pydict()["columns_mean"] == [4.0, 5.0, 6.0]
    assert df.to_pydict()["columns_avg"] == [4.0, 5.0, 6.0]


def test_columns_mean_with_no_input_columns(make_df) -> None:
    df = make_df({"a": [1, 2, 3]})

    with pytest.raises(ValueError, match="columns_mean requires at least one expression"):
        df.with_column("mean", columns_mean()).collect()

    with pytest.raises(ValueError, match="columns_avg requires at least one expression"):
        df.with_column("avg", columns_avg()).collect()


def test_columns_min(make_df) -> None:
    data = {"a": [1, 2, 3], "b": [4, 5, 6], "c": [7, 8, 9]}
    df = make_df(data).with_column("min", columns_min("a", "b", "c"))

    assert df.to_pydict()["min"] == [1, 2, 3]


def test_columns_min_with_nulls(make_df) -> None:
    data = {"a": [1, None, 3], "b": [4, 5, None], "c": [None, 8, 9]}
    df = make_df(data).with_column("min", columns_min("a", "b", "c"))

    assert df.to_pydict()["min"] == [1, 5, 3]


def test_columns_min_with_expressions(make_df) -> None:
    data = {"a": [1, 2, 3], "b": [1, 2, 3], "c": [1, 2, 3]}
    df = make_df(data).with_column("min", columns_min(col("a") + 1, col("b") + 2, col("c") + 3))

    assert df.to_pydict()["min"] == [2, 3, 4]


def test_columns_min_in_select(make_df) -> None:
    df = make_df({"a": [1, 2, 3], "b": [4, 5, 6], "c": [7, 8, 9]})
    df = df.select(columns_min("a", "b", "c"))

    assert df.to_pydict()["columns_min"] == [1, 2, 3]


def test_columns_min_with_no_input_columns(make_df) -> None:
    df = make_df({"a": [1, 2, 3]})

    with pytest.raises(ValueError, match="columns_min requires at least one expression"):
        df.with_column("min", columns_min()).collect()


def test_columns_max(make_df) -> None:
    data = {"a": [1, 2, 3], "b": [4, 5, 6], "c": [7, 8, 9]}
    df = make_df(data).with_column("max", columns_max("a", "b", "c"))

    assert df.to_pydict()["max"] == [7, 8, 9]


def test_columns_max_with_nulls(make_df) -> None:
    data = {"a": [1, None, 3], "b": [4, 5, None], "c": [None, 8, 9]}
    df = make_df(data).with_column("max", columns_max("a", "b", "c"))

    assert df.to_pydict()["max"] == [4, 8, 9]


def test_columns_max_with_expressions(make_df) -> None:
    data = {"a": [1, 2, 3], "b": [1, 2, 3], "c": [1, 2, 3]}
    df = make_df(data).with_column("max", columns_max(col("a") + 1, col("b") + 2, col("c") + 3))

    assert df.to_pydict()["max"] == [4, 5, 6]


def test_columns_max_in_select(make_df) -> None:
    df = make_df({"a": [1, 2, 3], "b": [4, 5, 6], "c": [7, 8, 9]})
    df = df.select(columns_max("a", "b", "c"))

    assert df.to_pydict()["columns_max"] == [7, 8, 9]


def test_columns_max_with_no_input_columns(make_df) -> None:
    df = make_df({"a": [1, 2, 3]})

    with pytest.raises(ValueError, match="columns_max requires at least one expression"):
        df.with_column("max", columns_max()).collect()



from __future__ import annotations

import pyarrow as pa
import pytest

from daft.daft import series_lit
from daft.expressions import col, lit
from daft.expressions.expressions import Expression
from daft.recordbatch import MicroPartition


@pytest.mark.parametrize(
    "data,value,expected_values",
    [
        ([(1, 10), (2, None)], 10, [True, False]),  # 10 <=> 10, NULL <=> 10
        ([(1, None), (2, None), (3, 10)], None, [True, True, False]),  # NULL <=> NULL, NULL <=> NULL, 10 <=> NULL
        ([(1, 10), (2, 20)], 10, [True, False]),  # 10 <=> 10, 20 <=> 10
    ],
)
def test_null_safe_equals_basic(data, value, expected_values):
    """Test basic null-safe equality comparison."""
    # Create a table with the test data
    table = MicroPartition.from_pydict(
        {
            "id": [x[0] for x in data],
            "value": [x[1] for x in data],
        }
    )

    # Apply the null-safe equals operation
    result = table.eval_expression_list([col("value").eq_null_safe(lit(value))])
    result_values = result.get_column_by_name("value").to_pylist()

    # Verify results
    assert result_values == expected_values


@pytest.mark.parametrize(
    "type_name,test_value,test_data,expected_values",
    [
        ("int", 10, [(1, 10), (2, None), (3, 20)], [True, False, False]),
        ("string", "hello", [(1, "hello"), (2, None), (3, "world")], [True, False, False]),
        ("boolean", True, [(1, True), (2, None), (3, False)], [True, False, False]),
        ("float", 1.5, [(1, 1.5), (2, None), (3, 2.5)], [True, False, False]),
        ("binary", b"hello", [(1, b"hello"), (2, None), (3, b"world")], [True, False, False]),
        ("fixed_size_binary", b"aaa", [(1, b"aaa"), (2, None), (3, b"bbb")], [True, False, False]),
        ("null", None, [(1, 10), (2, None), (3, 20)], [False, True, False]),
    ],
)
def test_null_safe_equals_types(type_name, test_value, test_data, expected_values):
    """Test null-safe equality with different data types."""
    # Create a table with the test data
    if type_name == "fixed_size_binary":
        # Convert to PyArrow array for fixed size binary
        value_array = pa.array([x[1] for x in test_data], type=pa.binary(3))
        table = MicroPartition.from_pydict(
            {
                "id": [x[0] for x in test_data],
                "value": value_array.to_pylist(),
            }
        )
    else:
        table = MicroPartition.from_pydict(
            {
                "id": [x[0] for x in test_data],
                "value": [x[1] for x in test_data],
            }
        )

    # Apply the null-safe equals operation
    result = table.eval_expression_list([col("value").eq_null_safe(lit(test_value))])
    result_values = result.get_column_by_name("value").to_pylist()

    # Verify results
    assert result_values == expected_values, f"Failed for {type_name} comparison"


@pytest.mark.parametrize(
    "data,expected_values",
    [
        ([(10, 10), (None, None), (10, None), (None, 10), (10, 20)], [True, True, False, False, False]),
        (
            [("hello", "hello"), (None, None), ("hello", None), (None, "hello"), ("hello", "world")],
            [True, True, False, False, False],
        ),
        ([(True, True), (None, None), (True, None), (None, True), (True, False)], [True, True, False, False, False]),
        ([(1.5, 1.5), (None, None), (1.5, None), (None, 1.5), (1.5, 2.5)], [True, True, False, False, False]),
        (
            [(b"hello", b"hello"), (None, None), (b"hello", None), (None, b"hello"), (b"hello", b"world")],
            [True, True, False, False, False],
        ),
        (
            [(b"aaa", b"aaa"), (None, None), (b"aaa", None), (None, b"aaa"), (b"aaa", b"bbb")],
            [True, True, False, False, False],
        ),
    ],
)
def test_null_safe_equals_column_comparison(data, expected_values):
    """Test null-safe equality between two columns."""
    # Check if this is a fixed-size binary test case
    is_fixed_binary = isinstance(data[0][0], bytes) and len(data[0][0]) == 3

    # Create a table with the test data
    if is_fixed_binary:
        # Convert to PyArrow arrays for fixed size binary
        left_array = pa.array([x[0] for x in data], type=pa.binary(3))
        right_array = pa.array([x[1] for x in data], type=pa.binary(3))
        table = MicroPartition.from_pydict(
            {
                "left": left_array.to_pylist(),
                "right": right_array.to_pylist(),
            }
        )
    else:
        table = MicroPartition.from_pydict(
            {
                "left": [x[0] for x in data],
                "right": [x[1] for x in data],
            }
        )

    # Apply the null-safe equals operation
    result = table.eval_expression_list([col("left").eq_null_safe(col("right"))])
    result_values = result.get_column_by_name("left").to_pylist()

    # Verify results
    assert result_values == expected_values


@pytest.mark.parametrize(
    "filter_value,data,expected_ids",
    [
        (10, [(1, 10), (2, None), (3, 20), (4, None)], {1}),  # Only id=1 has value=10
        (None, [(1, 10), (2, None), (3, 20), (4, None)], {2, 4}),  # id=2 and id=4 have NULL values
        (20, [(1, 10), (2, None), (3, 20), (4, None)], {3}),  # Only id=3 has value=20
    ],
)
def test_null_safe_equals_in_filter(filter_value, data, expected_ids):
    """Test using null-safe equality in filter."""
    # Create a table with the test data
    table = MicroPartition.from_pydict(
        {
            "id": [x[0] for x in data],
            "value": [x[1] for x in data],
        }
    )

    # Apply the filter with null-safe equals
    result = table.filter([col("value").eq_null_safe(lit(filter_value))])
    result_ids = set(result.get_column_by_name("id").to_pylist())

    # Verify results
    assert result_ids == expected_ids


@pytest.mark.parametrize(
    "operation,test_value,data,expected_values",
    [
        (
            "NOT",
            10,
            [(1, 10), (2, None), (3, 20)],
            [False, True, True],  # NOT (10 <=> 10), NOT (NULL <=> 10), NOT (20 <=> 10)
        ),
    ],
)
def test_null_safe_equals_chained_operations(operation, test_value, data, expected_values):
    """Test chaining null-safe equality with other operations."""
    # Create a table with the test data
    table = MicroPartition.from_pydict(
        {
            "id": [x[0] for x in data],
            "value": [x[1] for x in data],
        }
    )

    # Apply the operation
    if operation == "NOT":
        result = table.eval_expression_list([~col("value").eq_null_safe(lit(test_value))])

    result_values = result.get_column_by_name("value").to_pylist()

    # Verify results
    assert result_values == expected_values


def test_null_safe_equals_fixed_size_binary():
    """Test null-safe equality specifically for fixed-size binary arrays."""
    # Create arrays with fixed size binary data
    l_arrow = pa.array([b"11111", b"22222", b"33333", None, b"12345", None], type=pa.binary(5))
    r_arrow = pa.array([b"11111", b"33333", b"11111", b"12345", None, None], type=pa.binary(5))

    # Create table with these arrays
    table = MicroPartition.from_pydict(
        {
            "left": l_arrow.to_pylist(),
            "right": r_arrow.to_pylist(),
        }
    )

    # Test column to column comparison
    result = table.eval_expression_list([col("left").eq_null_safe(col("right"))])
    result_values = result.get_column_by_name("left").to_pylist()
    assert result_values == [True, False, False, False, False, True]  # True for equal values and both-null cases

    # Test column to scalar comparisons
    test_cases = [
        (b"11111", [True, False, False, False, False, False]),  # Matches first value
        (b"22222", [False, True, False, False, False, False]),  # Matches second value
        (None, [False, False, False, True, False, True]),  # True for all null values
    ]

    for test_value, expected in test_cases:
        result = table.eval_expression_list([col("left").eq_null_safe(lit(test_value))])
        result_values = result.get_column_by_name("left").to_pylist()
        assert result_values == expected, f"Failed for test value: {test_value}"

        # Test the reverse comparison as well
        result = table.eval_expression_list([col("right").eq_null_safe(lit(test_value))])
        result_values = result.get_column_by_name("right").to_pylist()
        expected_reverse = [
            True if test_value == b"11111" else False,  # First value is "11111"
            False,  # Second value is "33333"
            True if test_value == b"11111" else False,  # Third value is "11111"
            True if test_value == b"12345" else False,  # Fourth value is "12345"
            True if test_value is None else False,  # Fifth value is None
            True if test_value is None else False,  # Sixth value is None
        ]
        assert result_values == expected_reverse, f"Failed for reverse test value: {test_value}"


@pytest.mark.parametrize(
    "type_name,left_data,right_data,expected_values",
    [
        ("int", [1, 2, 3], [1, 2, 4], [True, True, False]),
        ("float", [1.0, 2.0, 3.0], [1.0, 2.0, 4.0], [True, True, False]),
        ("boolean", [True, False, True], [True, False, False], [True, True, False]),
        ("string", ["a", "b", "c"], ["a", "b", "d"], [True, True, False]),
        ("binary", [b"a", b"b", b"c"], [b"a", b"b", b"d"], [True, True, False]),
        ("fixed_size_binary", [b"aaa", b"bbb", b"ccc"], [b"aaa", b"bbb", b"ddd"], [True, True, False]),
    ],
)
def test_no_nulls_all_types(type_name, left_data, right_data, expected_values):
    """Test null-safe equality with no nulls in either array for all data types."""
    if type_name == "fixed_size_binary":
        # Convert to PyArrow arrays for fixed size binary
        left_array = pa.array(left_data, type=pa.binary(3))
        right_array = pa.array(right_data, type=pa.binary(3))
        table = MicroPartition.from_pydict(
            {
                "left": left_array.to_pylist(),
                "right": right_array.to_pylist(),
            }
        )
    else:
        table = MicroPartition.from_pydict(
            {
                "left": left_data,
                "right": right_data,
            }
        )

    result = table.eval_expression_list([col("left").eq_null_safe(col("right"))])
    result_values = result.get_column_by_name("left").to_pylist()

    assert result_values == expected_values, f"Failed for {type_name} comparison"


@pytest.mark.parametrize(
    "type_name,left_data,right_data,expected_values",
    [
        ("int", [1, 2, 3], [1, None, 3], [True, False, True]),
        ("float", [1.0, 2.0, 3.0], [1.0, None, 3.0], [True, False, True]),
        ("boolean", [True, False, True], [True, None, True], [True, False, True]),
        ("string", ["a", "b", "c"], ["a", None, "c"], [True, False, True]),
        ("binary", [b"a", b"b", b"c"], [b"a", None, b"c"], [True, False, True]),
        ("fixed_size_binary", [b"aaa", b"bbb", b"ccc"], [b"aaa", None, b"ccc"], [True, False, True]),
    ],
)
def test_right_nulls_all_types(type_name, left_data, right_data, expected_values):
    """Test null-safe equality where left array has no nulls and right array has some nulls."""
    if type_name == "fixed_size_binary":
        # Convert to PyArrow arrays for fixed size binary
        left_array = pa.array(left_data, type=pa.binary(3))
        right_array = pa.array(right_data, type=pa.binary(3))
        table = MicroPartition.from_pydict(
            {
                "left": left_array.to_pylist(),
                "right": right_array.to_pylist(),
            }
        )
    else:
        table = MicroPartition.from_pydict(
            {
                "left": left_data,
                "right": right_data,
            }
        )

    result = table.eval_expression_list([col("left").eq_null_safe(col("right"))])
    result_values = result.get_column_by_name("left").to_pylist()

    assert result_values == expected_values, f"Failed for {type_name} comparison"


@pytest.mark.parametrize(
    "type_name,left_data,right_data,expected_values",
    [
        ("int", [1, None, 3], [1, 2, 3], [True, False, True]),
        ("float", [1.0, None, 3.0], [1.0, 2.0, 3.0], [True, False, True]),
        ("boolean", [True, None, True], [True, False, True], [True, False, True]),
        ("string", ["a", None, "c"], ["a", "b", "c"], [True, False, True]),
        ("binary", [b"a", None, b"c"], [b"a", b"b", b"c"], [True, False, True]),
        ("fixed_size_binary", [b"aaa", None, b"ccc"], [b"aaa", b"bbb", b"ccc"], [True, False, True]),
    ],
)
def test_left_nulls_all_types(type_name, left_data, right_data, expected_values):
    """Test null-safe equality where left array has some nulls and right array has no nulls."""
    if type_name == "fixed_size_binary":
        # Convert to PyArrow arrays for fixed size binary
        left_array = pa.array(left_data, type=pa.binary(3))
        right_array = pa.array(right_data, type=pa.binary(3))
        table = MicroPartition.from_pydict(
            {
                "left": left_array.to_pylist(),
                "right": right_array.to_pylist(),
            }
        )
    else:
        table = MicroPartition.from_pydict(
            {
                "left": left_data,
                "right": right_data,
            }
        )

    result = table.eval_expression_list([col("left").eq_null_safe(col("right"))])
    result_values = result.get_column_by_name("left").to_pylist()

    assert result_values == expected_values, f"Failed for {type_name} comparison"


@pytest.mark.parametrize(
    "type_name,left_data,right_data",
    [
        ("int", [1, 2, 3], [1, 2]),
        ("float", [1.0, 2.0, 3.0], [1.0, 2.0]),
        ("boolean", [True, False, True], [True, False]),
        ("string", ["a", "b", "c"], ["a", "b"]),
        ("binary", [b"a", b"b", b"c"], [b"a", b"b"]),
        ("fixed_size_binary", [b"aaa", b"bbb", b"ccc"], [b"aaa", b"bbb"]),
    ],
)
def test_length_mismatch_all_types(type_name, left_data, right_data):
    """Test that length mismatches raise appropriate error for all data types."""
    # Create two separate tables
    left_table = MicroPartition.from_pydict({"value": left_data})
    right_table = MicroPartition.from_pydict({"value": right_data})

    with pytest.raises(ValueError) as exc_info:
        result = left_table.eval_expression_list(
            [
                col("value").eq_null_safe(
                    Expression._from_pyexpr(series_lit(right_table.get_column_by_name("value")._series))
                )
            ]
        )
        # Force evaluation by accessing the result
        result.get_column_by_name("value").to_pylist()

    # Verify error message format
    error_msg = str(exc_info.value)
    assert "trying to compare different length arrays" in error_msg


@pytest.mark.parametrize(
    "type_name,left_data,right_data,expected_values",
    [
        ("int", [None, None, 1], [None, None, 2], [True, True, False]),
        ("float", [None, None, 1.0], [None, None, 2.0], [True, True, False]),
        ("boolean", [None, None, True], [None, None, False], [True, True, False]),
        ("string", [None, None, "a"], [None, None, "b"], [True, True, False]),
        ("binary", [None, None, b"a"], [None, None, b"b"], [True, True, False]),
        ("fixed_size_binary", [None, None, b"aaa"], [None, None, b"bbb"], [True, True, False]),
    ],
)
def test_null_equals_null_all_types(type_name, left_data, right_data, expected_values):
    """Test that NULL <=> NULL returns True for all data types."""
    if type_name == "fixed_size_binary":
        # Convert to PyArrow arrays for fixed size binary
        left_array = pa.array(left_data, type=pa.binary(3))
        right_array = pa.array(right_data, type=pa.binary(3))
        table = MicroPartition.from_pydict(
            {
                "left": left_array.to_pylist(),
                "right": right_array.to_pylist(),
            }
        )
    else:
        table = MicroPartition.from_pydict(
            {
                "left": left_data,
                "right": right_data,
            }
        )

    result = table.eval_expression_list([col("left").eq_null_safe(col("right"))])
    result_values = result.get_column_by_name("left").to_pylist()

    assert result_values == expected_values, f"Failed for {type_name} comparison"



from __future__ import annotations

from typing import Any

import pytest

import daft
from tests.conftest import check_answer


def make_spark_df(spark, data: dict[str, Any]):
    fields = [name for name in data]
    rows = list(zip(*[data[name] for name in fields]))
    return spark.createDataFrame(rows, fields)


def helper(spark, make_df, op: str, left: dict[str, Any], right: dict[str, Any], expected: dict[str, Any]):
    df1 = make_spark_df(spark, left)
    df2 = make_spark_df(spark, right)
    df_helper(op, df1, df2, expected)


def df_helper(op: str, df1, df2, expected: dict[str, Any]):
    if op == "intersect":
        result = df1.intersect(df2)
    elif op == "intersect_all":
        result = df1.intersectAll(df2)
    else:
        result = df1.exceptAll(df2)
    result = daft.from_pandas(result.toPandas())
    check_answer(result, expected)


@pytest.mark.parametrize(
    "op, left, right, expected",
    [
        ("intersect", {"foo": [1, 2, 3]}, {"bar": [2, 3, 4]}, {"foo": [2, 3]}),
        ("intersect_all", {"foo": [1, 2, 2]}, {"bar": [2, 2, 4]}, {"foo": [2, 2]}),
        ("except_all", {"foo": [1, 2, 2]}, {"bar": [2, 4]}, {"foo": [1, 2]}),
    ],
)
def test_simple_intersect_or_except(spark_session, make_df, op, left, right, expected):
    helper(spark_session, make_df, op, left, right, expected)


@pytest.mark.parametrize(
    "op, left, right, expected",
    [
        ("intersect", {"foo": [1, 2, 2, 3]}, {"bar": [2, 3, 3]}, {"foo": [2, 3]}),
        ("intersect_all", {"foo": [1, 2, 2, 3]}, {"bar": [2, 3, 3]}, {"foo": [2, 3]}),
        ("except_all", {"foo": [1, 2, 2, 3]}, {"bar": [2, 3, 3]}, {"foo": [1, 2]}),
    ],
)
def test_with_duplicate(spark_session, make_df, op, left, right, expected):
    helper(spark_session, make_df, op, left, right, expected)


@pytest.mark.parametrize(
    "op, df, expected",
    [
        ("intersect", {"foo": [1, 2, 3]}, {"foo": [1, 2, 3]}),
        ("intersect_all", {"foo": [1, 2, 3]}, {"foo": [1, 2, 3]}),
        ("except_all", {"foo": [1, 2, 2]}, {"foo": []}),
    ],
)
def test_with_self(spark_session, make_df, op, df, expected):
    df = make_spark_df(spark_session, df)
    df_helper(op, df, df, expected)


@pytest.mark.parametrize(
    "op, left, right, expected",
    [
        ("intersect", {"foo": [1, 2, None]}, {"foo": [2, 3, None]}, {"foo": [2, None]}),
        ("intersect_all", {"foo": [1, 2, None]}, {"foo": [2, 3, None]}, {"foo": [2, None]}),
        ("intersect", {"foo": [1, 2]}, {"foo": [2, 3, None]}, {"foo": [2]}),
        ("intersect_all", {"foo": [1, 2]}, {"foo": [2, 3, None]}, {"foo": [2]}),
        ("intersect", {"foo": [1, 2, None]}, {"foo": [2, 3]}, {"foo": [2]}),
        ("intersect_all", {"foo": [1, 2, None]}, {"foo": [2, 3]}, {"foo": [2]}),
    ],
)
def test_intersect_with_nulls(spark_session, make_df, op, left, right, expected):
    helper(spark_session, make_df, op, left, right, expected)


@pytest.mark.parametrize(
    "op, left, right, expected",
    [
        ("except_all", {"foo": [1, 2, None]}, {"foo": [2, 3, None]}, {"foo": [1]}),
        ("except_all", {"foo": [1, 2]}, {"foo": [2, 3, None]}, {"foo": [1]}),
        ("except_all", {"foo": [1, 2, None]}, {"foo": [2, 3]}, {"foo": [1, None]}),
    ],
)
def test_except_with_nulls(spark_session, make_df, op, left, right, expected):
    helper(spark_session, make_df, op, left, right, expected)


@pytest.mark.parametrize(
    "op, left, right, expected",
    [
        (
            "intersect_all",
            {"foo": [1, 2, 2], "bar": [2, 3, 3]},
            {"a": [2, 2, 4], "b": [3, 3, 4]},
            {"foo": [2, 2], "bar": [3, 3]},
        ),
        (
            "except_all",
            {"foo": [1, 2, 2], "bar": [2, 3, 3]},
            {"a": [2, 2, 4], "b": [3, 3, 4]},
            {"foo": [1], "bar": [2]},
        ),
    ],
)
def test_multiple_fields(spark_session, make_df, op, left, right, expected):
    helper(spark_session, make_df, op, left, right, expected)


def test_union(spark_session, spark_to_daft):
    df1 = spark_session.createDataFrame([(1, "A"), (2, "B")], ["id", "value"])
    df2 = spark_session.createDataFrame([(3, "C"), (4, "D")], ["id", "value"])
    df3 = df1.union(df2)
    actual = spark_to_daft(df3)
    expected = {"id": [1, 2, 3, 4], "value": ["A", "B", "C", "D"]}

    assert actual.to_pydict() == expected


def test_union_by_name(spark_session, spark_to_daft):
    df1 = spark_session.createDataFrame([[1, 2, 3]], ["col0", "col1", "col2"])
    df2 = spark_session.createDataFrame([[4, 5, 6]], ["col1", "col2", "col0"])
    actual = spark_to_daft(df1.unionByName(df2))
    expected = {"col0": [1, 6], "col1": [2, 4], "col2": [3, 5]}
    assert actual.to_pydict() == expected



from __future__ import annotations

import itertools

import numpy as np
import pyarrow as pa
import pytest

from daft import col
from daft.logical.schema import Schema
from daft.recordbatch import MicroPartition
from daft.series import Series
from tests.recordbatch import daft_numeric_types, daft_string_types


@pytest.mark.parametrize(
    "mp",
    [
        MicroPartition.from_pydict({"a": pa.array([], type=pa.int64())}),  # 1 empty table
        MicroPartition.empty(Schema.from_pyarrow_schema(pa.schema({"a": pa.int64()}))),  # No tables
    ],
)
def test_micropartitions_sort_empty(mp) -> None:
    sorted_table = mp.sort([col("a")])
    assert len(mp) == len(sorted_table) == 0
    assert mp.schema() == sorted_table.schema()


@pytest.mark.parametrize(
    "mp",
    [
        MicroPartition.from_pydict({"a": [1, 3, 2, 4]}),  # 1 table
        MicroPartition.concat(
            [
                MicroPartition.from_pydict({"a": np.array([]).astype(np.int64)}),
                MicroPartition.from_pydict({"a": [1]}),
                MicroPartition.from_pydict({"a": [3, 2, 4]}),
            ]
        ),  # 3 tables
    ],
)
def test_micropartitions_sort(mp) -> None:
    sorted_table = mp.sort([col("a")])
    assert len(mp) == len(sorted_table) == 4
    assert mp.schema() == sorted_table.schema()
    assert sorted_table.to_pydict()["a"] == sorted(mp.to_pydict()["a"])


@pytest.mark.parametrize(
    "sort_dtype, value_dtype, first_col",
    itertools.product(daft_numeric_types + daft_string_types, daft_numeric_types + daft_string_types, [False, True]),
)
def test_table_single_col_sorting(sort_dtype, value_dtype, first_col) -> None:
    pa_table = pa.Table.from_pydict({"a": [None, 4, 2, 1, 5], "b": [0, 1, 2, 3, None]})

    argsort_order = Series.from_pylist([3, 2, 1, 4, 0])

    daft_recordbatch = MicroPartition.from_arrow(pa_table)

    if first_col:
        daft_recordbatch = daft_recordbatch.eval_expression_list(
            [col("a").cast(sort_dtype), col("b").cast(value_dtype)]
        )
    else:
        daft_recordbatch = daft_recordbatch.eval_expression_list(
            [col("b").cast(value_dtype), col("a").cast(sort_dtype)]
        )

    assert len(daft_recordbatch) == 5
    if first_col:
        assert daft_recordbatch.column_names() == ["a", "b"]
    else:
        assert daft_recordbatch.column_names() == ["b", "a"]

    sorted_table = daft_recordbatch.sort([col("a")])

    assert len(sorted_table) == 5

    if first_col:
        assert sorted_table.column_names() == ["a", "b"]
    else:
        assert sorted_table.column_names() == ["b", "a"]

    assert sorted_table.get_column_by_name("a").datatype() == daft_recordbatch.get_column_by_name("a").datatype()
    assert sorted_table.get_column_by_name("b").datatype() == daft_recordbatch.get_column_by_name("b").datatype()

    assert (
        sorted_table.get_column_by_name("a").to_pylist()
        == daft_recordbatch.get_column_by_name("a").take(argsort_order).to_pylist()
    )
    assert (
        sorted_table.get_column_by_name("b").to_pylist()
        == daft_recordbatch.get_column_by_name("b").take(argsort_order).to_pylist()
    )

    assert daft_recordbatch.argsort([col("a")]).to_pylist() == argsort_order.to_pylist()

    # Descending

    sorted_table = daft_recordbatch.sort([col("a")], descending=True)

    assert len(sorted_table) == 5
    if first_col:
        assert sorted_table.column_names() == ["a", "b"]
    else:
        assert sorted_table.column_names() == ["b", "a"]

    assert sorted_table.get_column_by_name("a").datatype() == daft_recordbatch.get_column_by_name("a").datatype()
    assert sorted_table.get_column_by_name("b").datatype() == daft_recordbatch.get_column_by_name("b").datatype()

    assert (
        sorted_table.get_column_by_name("a").to_pylist()
        == daft_recordbatch.get_column_by_name("a").take(argsort_order).to_pylist()[::-1]
    )
    assert (
        sorted_table.get_column_by_name("b").to_pylist()
        == daft_recordbatch.get_column_by_name("b").take(argsort_order).to_pylist()[::-1]
    )

    assert daft_recordbatch.argsort([col("a")], descending=True).to_pylist() == argsort_order.to_pylist()[::-1]


@pytest.mark.parametrize(
    "sort_dtype, value_dtype, data",
    itertools.product(
        daft_numeric_types + daft_string_types,
        daft_numeric_types + daft_string_types,
        [
            ([None, 4, 2, 1, 5], [0, 1, 2, 3, None], False, False, [3, 2, 1, 4, 0]),
            ([None, 4, 2, 1, 5], [0, 1, 2, 3, None], False, True, [3, 2, 1, 4, 0]),
            ([1, 1, 1, 1, 1], [None, 3, 1, 2, 0], False, False, [4, 2, 3, 1, 0]),
            ([1, 1, 1, 1, 1], [None, 3, 1, 2, 0], True, False, [4, 2, 3, 1, 0]),
            ([None, None, None, None, None], [None, 3, 1, 2, 0], False, False, [4, 2, 3, 1, 0]),
            ([None, None, None, None, None], [None, 3, 1, 2, 0], True, False, [4, 2, 3, 1, 0]),
            ([None, 4, 2, 1, 5], [None, None, None, None, None], False, False, [3, 2, 1, 4, 0]),
            ([None, 4, 2, 1, 5], [None, None, None, None, None], False, True, [3, 2, 1, 4, 0]),
        ],
    ),
)
def test_table_multiple_col_sorting(sort_dtype, value_dtype, data) -> None:
    a, b, a_desc, b_desc, expected = data
    pa_table = pa.Table.from_pydict({"a": a, "b": b})

    argsort_order = Series.from_pylist(expected)

    daft_recordbatch = MicroPartition.from_arrow(pa_table)

    daft_recordbatch = daft_recordbatch.eval_expression_list([col("a").cast(sort_dtype), col("b").cast(value_dtype)])

    assert len(daft_recordbatch) == 5
    assert daft_recordbatch.column_names() == ["a", "b"]

    sorted_table = daft_recordbatch.sort([col("a"), col("b")], descending=[a_desc, b_desc])

    assert len(sorted_table) == 5

    assert sorted_table.column_names() == ["a", "b"]

    assert sorted_table.get_column_by_name("a").datatype() == daft_recordbatch.get_column_by_name("a").datatype()
    assert sorted_table.get_column_by_name("b").datatype() == daft_recordbatch.get_column_by_name("b").datatype()

    assert (
        sorted_table.get_column_by_name("a").to_pylist()
        == daft_recordbatch.get_column_by_name("a").take(argsort_order).to_pylist()
    )
    assert (
        sorted_table.get_column_by_name("b").to_pylist()
        == daft_recordbatch.get_column_by_name("b").take(argsort_order).to_pylist()
    )

    assert (
        daft_recordbatch.argsort([col("a"), col("b")], descending=[a_desc, b_desc]).to_pylist()
        == argsort_order.to_pylist()
    )

    # Descending

    sorted_table = daft_recordbatch.sort([col("a"), col("b")], descending=[not a_desc, not b_desc])

    assert len(sorted_table) == 5

    assert sorted_table.column_names() == ["a", "b"]

    assert sorted_table.get_column_by_name("a").datatype() == daft_recordbatch.get_column_by_name("a").datatype()
    assert sorted_table.get_column_by_name("b").datatype() == daft_recordbatch.get_column_by_name("b").datatype()

    assert (
        sorted_table.get_column_by_name("a").to_pylist()
        == daft_recordbatch.get_column_by_name("a").take(argsort_order).to_pylist()[::-1]
    )
    assert (
        sorted_table.get_column_by_name("b").to_pylist()
        == daft_recordbatch.get_column_by_name("b").take(argsort_order).to_pylist()[::-1]
    )

    assert (
        daft_recordbatch.argsort([col("a"), col("b")], descending=[not a_desc, not b_desc]).to_pylist()
        == argsort_order.to_pylist()[::-1]
    )


@pytest.mark.parametrize(
    "data",
    [
        ([None, 4, 2, 1, 5], [0, 1, 2, 3, None], False, False, [3, 2, 1, 4, 0]),
        ([None, 4, 2, 1, 5], [0, 1, 2, 3, None], False, True, [3, 2, 1, 4, 0]),
        ([1, 1, 1, 1, 1], [None, 3, 1, 2, 0], False, False, [4, 2, 3, 1, 0]),
        ([1, 1, 1, 1, 1], [None, 3, 1, 2, 0], True, False, [4, 2, 3, 1, 0]),
        ([None, None, None, None, None], [None, 3, 1, 2, 0], False, False, [4, 2, 3, 1, 0]),
        ([None, None, None, None, None], [None, 3, 1, 2, 0], True, False, [4, 2, 3, 1, 0]),
        ([None, 4, 2, 1, 5], [None, None, None, None, None], False, False, [3, 2, 1, 4, 0]),
        ([None, 4, 2, 1, 5], [None, None, None, None, None], False, True, [3, 2, 1, 4, 0]),
    ],
)
def test_table_multiple_col_sorting_binary(data) -> None:
    a, b, a_desc, b_desc, expected = data
    a = pa.array([x.to_bytes(1, "little") if x is not None else None for x in a], type=pa.binary())
    b = pa.array([x.to_bytes(1, "little") if x is not None else None for x in b], type=pa.binary())

    pa_table = pa.Table.from_pydict({"a": a, "b": b})

    argsort_order = Series.from_pylist(expected)

    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 5
    assert daft_recordbatch.column_names() == ["a", "b"]

    sorted_table = daft_recordbatch.sort([col("a"), col("b")], descending=[a_desc, b_desc])

    assert len(sorted_table) == 5

    assert sorted_table.column_names() == ["a", "b"]

    assert sorted_table.get_column_by_name("a").datatype() == daft_recordbatch.get_column_by_name("a").datatype()
    assert sorted_table.get_column_by_name("b").datatype() == daft_recordbatch.get_column_by_name("b").datatype()

    assert (
        sorted_table.get_column_by_name("a").to_pylist()
        == daft_recordbatch.get_column_by_name("a").take(argsort_order).to_pylist()
    )
    assert (
        sorted_table.get_column_by_name("b").to_pylist()
        == daft_recordbatch.get_column_by_name("b").take(argsort_order).to_pylist()
    )

    assert (
        daft_recordbatch.argsort([col("a"), col("b")], descending=[a_desc, b_desc]).to_pylist()
        == argsort_order.to_pylist()
    )

    # Descending

    sorted_table = daft_recordbatch.sort([col("a"), col("b")], descending=[not a_desc, not b_desc])

    assert len(sorted_table) == 5

    assert sorted_table.column_names() == ["a", "b"]

    assert sorted_table.get_column_by_name("a").datatype() == daft_recordbatch.get_column_by_name("a").datatype()
    assert sorted_table.get_column_by_name("b").datatype() == daft_recordbatch.get_column_by_name("b").datatype()

    assert (
        sorted_table.get_column_by_name("a").to_pylist()
        == daft_recordbatch.get_column_by_name("a").take(argsort_order).to_pylist()[::-1]
    )
    assert (
        sorted_table.get_column_by_name("b").to_pylist()
        == daft_recordbatch.get_column_by_name("b").take(argsort_order).to_pylist()[::-1]
    )

    assert (
        daft_recordbatch.argsort([col("a"), col("b")], descending=[not a_desc, not b_desc]).to_pylist()
        == argsort_order.to_pylist()[::-1]
    )


@pytest.mark.parametrize(
    "second_dtype, data",
    itertools.product(
        daft_numeric_types + daft_string_types,
        [
            ([None, True, False, True, False], [0, 1, 2, 3, None], False, False, [2, 4, 1, 3, 0]),
            ([None, True, False, True, False], [0, 1, 2, 3, None], True, False, [0, 1, 3, 2, 4]),
            ([True, True, True, True, True], [None, 3, 1, 2, 0], False, False, [4, 2, 3, 1, 0]),
            ([True, True, True, True, True], [None, 3, 1, 2, 0], True, False, [4, 2, 3, 1, 0]),
            ([None, None, None, None, None], [None, 3, 1, 2, 0], False, False, [4, 2, 3, 1, 0]),
            ([None, None, None, None, None], [None, 3, 1, 2, 0], True, False, [4, 2, 3, 1, 0]),
        ],
    ),
)
def test_table_boolean_multiple_col_sorting(second_dtype, data) -> None:
    a, b, a_desc, b_desc, expected = data
    pa_table = pa.Table.from_pydict({"a": a, "b": b})
    argsort_order = Series.from_pylist(expected)

    daft_recordbatch = MicroPartition.from_arrow(pa_table)

    daft_recordbatch = daft_recordbatch.eval_expression_list([col("a"), col("b").cast(second_dtype)])

    assert len(daft_recordbatch) == 5
    assert daft_recordbatch.column_names() == ["a", "b"]

    sorted_table = daft_recordbatch.sort([col("a"), col("b")], descending=[a_desc, b_desc])

    assert len(sorted_table) == 5

    assert sorted_table.column_names() == ["a", "b"]

    assert sorted_table.get_column_by_name("a").datatype() == daft_recordbatch.get_column_by_name("a").datatype()
    assert sorted_table.get_column_by_name("b").datatype() == daft_recordbatch.get_column_by_name("b").datatype()

    assert (
        sorted_table.get_column_by_name("a").to_pylist()
        == daft_recordbatch.get_column_by_name("a").take(argsort_order).to_pylist()
    )
    assert (
        sorted_table.get_column_by_name("b").to_pylist()
        == daft_recordbatch.get_column_by_name("b").take(argsort_order).to_pylist()
    )

    assert (
        daft_recordbatch.argsort([col("a"), col("b")], descending=[a_desc, b_desc]).to_pylist()
        == argsort_order.to_pylist()
    )

    # Descending

    sorted_table = daft_recordbatch.sort([col("a"), col("b")], descending=[not a_desc, not b_desc])

    assert len(sorted_table) == 5

    assert sorted_table.column_names() == ["a", "b"]

    assert sorted_table.get_column_by_name("a").datatype() == daft_recordbatch.get_column_by_name("a").datatype()
    assert sorted_table.get_column_by_name("b").datatype() == daft_recordbatch.get_column_by_name("b").datatype()

    assert (
        sorted_table.get_column_by_name("a").to_pylist()
        == daft_recordbatch.get_column_by_name("a").take(argsort_order).to_pylist()[::-1]
    )
    assert (
        sorted_table.get_column_by_name("b").to_pylist()
        == daft_recordbatch.get_column_by_name("b").take(argsort_order).to_pylist()[::-1]
    )

    assert (
        daft_recordbatch.argsort([col("a"), col("b")], descending=[not a_desc, not b_desc]).to_pylist()
        == argsort_order.to_pylist()[::-1]
    )


def test_table_sample() -> None:
    pa_table = pa.Table.from_pydict({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    source_pairs = {(1, 5), (2, 6), (3, 7), (4, 8)}

    daft_recordbatch = MicroPartition.from_arrow(pa_table)
    assert len(daft_recordbatch) == 4
    assert daft_recordbatch.column_names() == ["a", "b"]

    # subsample
    sampled = daft_recordbatch.sample(size=3)
    assert len(sampled) == 3
    assert sampled.column_names() == ["a", "b"]
    assert all(
        pair in source_pairs
        for pair in zip(sampled.get_column_by_name("a").to_pylist(), sampled.get_column_by_name("b").to_pylist())
    )

    # oversample
    sampled = daft_recordbatch.sample(size=4)
    assert len(sampled) == 4
    assert sampled.column_names() == ["a", "b"]
    assert all(
        pair in source_pairs
        for pair in zip(sampled.get_column_by_name("a").to_pylist(), sampled.get_column_by_name("b").to_pylist())
    )

    # negative sample
    with pytest.raises(ValueError, match="negative size"):
        daft_recordbatch.sample(size=-1)

    # fraction > 1.0
    with pytest.raises(ValueError, match="fraction greater than 1.0"):
        daft_recordbatch.sample(fraction=1.1)

    # fraction < 0.0
    with pytest.raises(ValueError, match="negative fraction"):
        daft_recordbatch.sample(fraction=-0.1)

    # size and fraction
    with pytest.raises(ValueError, match="Must specify either `fraction` or `size`"):
        daft_recordbatch.sample(size=1, fraction=0.5)

    # no arguments
    with pytest.raises(ValueError, match="Must specify either `fraction` or `size`"):
        daft_recordbatch.sample()


@pytest.mark.parametrize("size, k", itertools.product([0, 1, 10, 33, 100, 101], [0, 1, 2, 3, 100, 101, 200]))
def test_table_quantiles(size, k) -> None:
    first = np.arange(size)

    second = 2 * first

    daft_recordbatch = MicroPartition.from_pydict({"a": first, "b": second})
    assert len(daft_recordbatch) == size
    assert daft_recordbatch.column_names() == ["a", "b"]

    # sub
    quantiles = daft_recordbatch.quantiles(k)

    if size > 0:
        assert len(quantiles) == max(k - 1, 0)
    else:
        assert len(quantiles) == 0

    assert quantiles.column_names() == ["a", "b"]
    ind = quantiles.get_column_by_name("a").to_pylist()

    if k > 0:
        assert np.all(np.diff(ind) >= 0)
        expected_delta = size / k
        assert np.all(np.abs(np.diff(ind) - expected_delta) <= 1)
    else:
        assert len(ind) == 0


def test_table_quantiles_bad_input() -> None:
    # negative sample

    first = np.arange(10)

    second = 2 * first

    pa_table = pa.Table.from_pydict({"a": first, "b": second})

    daft_recordbatch = MicroPartition.from_arrow(pa_table)

    with pytest.raises(ValueError, match="negative number"):
        daft_recordbatch.quantiles(-1)


def test_string_table_sorting():
    daft_recordbatch = MicroPartition.from_pydict(
        {
            "firstname": [
                "bob",
                "alice",
                "eve",
                None,
                None,
                "bob",
                "alice",
            ],
            "lastname": ["a", "a", "a", "bond", None, None, "a"],
        }
    )
    sorted_table = daft_recordbatch.sort([col("firstname"), col("lastname")])
    assert sorted_table.to_pydict() == {
        "firstname": ["alice", "alice", "bob", "bob", "eve", None, None],
        "lastname": ["a", "a", "a", None, "a", "bond", None],
    }



from __future__ import annotations

from typing import Any

import pytest

import daft
from daft import DataFrame


def test_filter_missing_column(make_df, valid_data: list[dict[str, Any]]) -> None:
    df = make_df(valid_data)
    with pytest.raises(ValueError):
        df.select("sepal_width").where(df["petal_length"] > 4.8)


def test_drop_na(make_df, missing_value_data: list[dict[str, Any]]) -> None:
    df: DataFrame = make_df(missing_value_data)
    df_len_no_col = len(df.drop_nan().collect())
    assert df_len_no_col == 2

    df: DataFrame = make_df(missing_value_data)
    df_len_col = len(df.drop_nan("sepal_width").collect())
    assert df_len_col == 2


def test_drop_null(make_df, missing_value_data: list[dict[str, Any]]) -> None:
    df: DataFrame = make_df(missing_value_data)
    df_len_no_col = len(df.drop_null().collect())
    assert df_len_no_col == 2

    df: DataFrame = make_df(missing_value_data)
    df_len_col = len(df.drop_null("sepal_length").collect())
    assert df_len_col == 2


def test_filter_sql() -> None:
    df = daft.from_pydict({"x": [1, 2, 3], "y": [4, 5, 6], "z": [7, 9, 9]})

    df = df.where("z = 9 AND y > 5").collect().to_pydict()
    expected = {"x": [3], "y": [6], "z": [9]}

    assert df == expected


def test_filter_alias_for_where() -> None:
    df = daft.from_pydict({"x": [1, 2, 3], "y": [4, 5, 6], "z": [7, 9, 9]})

    expected = df.where("z = 9 AND y > 5").collect().to_pydict()
    actual = df.filter("z = 9 AND y > 5").collect().to_pydict()

    assert actual == expected



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Callable

from daft.daft import (
    CsvConvertOptions,
    CsvParseOptions,
    CsvReadOptions,
    IOConfig,
    JoinSide,
    JoinType,
    JsonConvertOptions,
    JsonParseOptions,
    JsonReadOptions,
)
from daft.daft import PyMicroPartition as _PyMicroPartition
from daft.daft import PyRecordBatch as _PyRecordBatch
from daft.daft import ScanTask as _ScanTask
from daft.datatype import DataType, TimeUnit
from daft.dependencies import pa
from daft.expressions import Expression, ExpressionsProjection
from daft.logical.schema import Schema
from daft.recordbatch.recordbatch import RecordBatch
from daft.series import Series

if TYPE_CHECKING:
    from collections.abc import Mapping

    import pandas as pd

logger = logging.getLogger(__name__)


class MicroPartition:
    _micropartition: _PyMicroPartition

    def __init__(self) -> None:
        raise NotImplementedError("We do not support creating a MicroPartition via __init__ ")

    def schema(self) -> Schema:
        return Schema._from_pyschema(self._micropartition.schema())

    def column_names(self) -> list[str]:
        return self._micropartition.column_names()

    def get_column_by_name(self, name: str) -> Series:
        return Series._from_pyseries(self._micropartition.get_column_by_name(name))

    def get_column(self, idx: int) -> Series:
        return Series._from_pyseries(self._micropartition.get_column(idx))

    def columns(self) -> list[Series]:
        return [Series._from_pyseries(s) for s in self._micropartition.columns()]

    def get_record_batches(self) -> list[RecordBatch]:
        return [RecordBatch._from_pyrecordbatch(t) for t in self._micropartition.get_record_batches()]

    def size_bytes(self) -> int | None:
        return self._micropartition.size_bytes()

    def __len__(self) -> int:
        return len(self._micropartition)

    def __repr__(self) -> str:
        return repr(self._micropartition)

    def _repr_html_(self) -> str:
        return self._micropartition._repr_html_()

    ###
    # Creation methods
    ###

    @staticmethod
    def empty(schema: Schema | None = None) -> MicroPartition:
        pyt = _PyMicroPartition.empty(None) if schema is None else _PyMicroPartition.empty(schema._schema)
        return MicroPartition._from_pymicropartition(pyt)

    @staticmethod
    def _from_scan_task(scan_task: _ScanTask) -> MicroPartition:
        assert isinstance(scan_task, _ScanTask)
        return MicroPartition._from_pymicropartition(_PyMicroPartition.from_scan_task(scan_task))

    @staticmethod
    def _from_pyrecordbatch(pyt: _PyRecordBatch) -> MicroPartition:
        assert isinstance(pyt, _PyRecordBatch)
        return MicroPartition._from_pymicropartition(_PyMicroPartition.from_record_batches([pyt]))

    @staticmethod
    def _from_pymicropartition(pym: _PyMicroPartition) -> MicroPartition:
        assert isinstance(pym, _PyMicroPartition)
        tab = MicroPartition.__new__(MicroPartition)
        tab._micropartition = pym
        return tab

    @staticmethod
    def _from_record_batches(tables: list[RecordBatch]) -> MicroPartition:
        return MicroPartition._from_pymicropartition(
            _PyMicroPartition.from_record_batches([t._recordbatch for t in tables])
        )

    @staticmethod
    def from_arrow(arrow_table: pa.Table) -> MicroPartition:
        record_batch = RecordBatch.from_arrow_table(arrow_table)
        return MicroPartition._from_record_batches([record_batch])

    @staticmethod
    def from_arrow_record_batches(rbs: list[pa.RecordBatch], arrow_schema: pa.Schema) -> MicroPartition:
        schema = Schema._from_field_name_and_types([(f.name, DataType.from_arrow_type(f.type)) for f in arrow_schema])
        pyt = _PyMicroPartition.from_arrow_record_batches(rbs, schema._schema)
        return MicroPartition._from_pymicropartition(pyt)

    @staticmethod
    def from_pandas(pd_df: pd.DataFrame) -> MicroPartition:
        table = RecordBatch.from_pandas(pd_df)
        return MicroPartition._from_record_batches([table])

    @staticmethod
    def from_pydict(data: Mapping[str, Any]) -> MicroPartition:
        table = RecordBatch.from_pydict(data)
        return MicroPartition._from_record_batches([table])

    @staticmethod
    def from_ipc_stream(stream: bytes) -> MicroPartition:
        return MicroPartition._from_pymicropartition(_PyMicroPartition.read_from_ipc_stream(stream))

    @classmethod
    def concat(cls, to_merge: list[MicroPartition]) -> MicroPartition:
        micropartitions = []
        for t in to_merge:
            if not isinstance(t, MicroPartition):
                raise TypeError(f"Expected a MicroPartition for concat, got {type(t)}")
            micropartitions.append(t._micropartition)
        return MicroPartition._from_pymicropartition(_PyMicroPartition.concat(micropartitions))

    def slice(self, start: int, end: int) -> MicroPartition:
        if not isinstance(start, int):
            raise TypeError(f"expected int for start but got {type(start)}")
        if not isinstance(end, int):
            raise TypeError(f"expected int for end but got {type(end)}")
        return MicroPartition._from_pymicropartition(self._micropartition.slice(start, end))

    ###
    # Exporting methods
    ###

    def to_table(self) -> RecordBatch:
        """DEPRECATED: Please use `to_record_batch`."""
        return self.to_record_batch()

    def to_record_batch(self) -> RecordBatch:
        """Returns the MicroPartition as a RecordBatch."""
        return RecordBatch._from_pyrecordbatch(self._micropartition.to_record_batch())

    def to_arrow(self, concat_record_batches: bool = False) -> pa.Table:
        if len(self) > 0 and not concat_record_batches:
            return pa.Table.from_batches(rb.to_arrow_record_batch() for rb in self.get_record_batches())
        else:
            return self.to_record_batch().to_arrow_table()

    def to_pydict(self) -> dict[str, list[Any]]:
        return self.to_record_batch().to_pydict()

    def to_pylist(self) -> list[dict[str, Any]]:
        return self.to_record_batch().to_pylist()

    def to_pandas(
        self,
        schema: Schema | None = None,
        coerce_temporal_nanoseconds: bool = False,
    ) -> pd.DataFrame:
        return self.to_record_batch().to_pandas(
            schema=schema,
            coerce_temporal_nanoseconds=coerce_temporal_nanoseconds,
        )

    def to_ipc_stream(self) -> bytes:
        return self._micropartition.write_to_ipc_stream()

    ###
    # Compute methods (MicroPartition -> MicroPartition)
    ###

    def eval_expression_list(self, exprs: ExpressionsProjection) -> MicroPartition:
        assert all(isinstance(e, Expression) for e in exprs)
        pyexprs = [e._expr for e in exprs]
        return MicroPartition._from_pymicropartition(self._micropartition.eval_expression_list(pyexprs))

    def head(self, num: int) -> MicroPartition:
        return MicroPartition._from_pymicropartition(self._micropartition.head(num))

    def take(self, indices: Series) -> MicroPartition:
        assert isinstance(indices, Series)
        return MicroPartition._from_pymicropartition(self._micropartition.take(indices._series))

    def filter(self, exprs: ExpressionsProjection) -> MicroPartition:
        assert all(isinstance(e, Expression) for e in exprs)
        pyexprs = [e._expr for e in exprs]
        return MicroPartition._from_pymicropartition(self._micropartition.filter(pyexprs))

    def sort(
        self,
        sort_keys: ExpressionsProjection,
        descending: bool | list[bool] | None = None,
        nulls_first: bool | list[bool] | None = None,
    ) -> MicroPartition:
        assert all(isinstance(e, Expression) for e in sort_keys)
        pyexprs = [e._expr for e in sort_keys]
        if descending is None:
            descending = [False for _ in pyexprs]
        elif isinstance(descending, bool):
            descending = [descending for _ in pyexprs]
        elif isinstance(descending, list):
            if len(descending) != len(sort_keys):
                raise ValueError(
                    f"Expected length of `descending` to be the same length as `sort_keys` since a list was passed in,"
                    f"got {len(descending)} instead of {len(sort_keys)}"
                )
        else:
            raise TypeError(f"Expected a bool, list[bool] or None for `descending` but got {type(descending)}")
        if nulls_first is None:
            nulls_first = descending
        elif isinstance(nulls_first, bool):
            nulls_first = [nulls_first for _ in pyexprs]
        elif isinstance(nulls_first, list):
            if len(nulls_first) != len(sort_keys):
                raise ValueError(
                    f"Expected length of `nulls_first` to be the same length as `sort_keys` since a list was passed in,"
                    f"got {len(nulls_first)} instead of {len(sort_keys)}"
                )
            else:
                nulls_first = [bool(x) for x in nulls_first]
        else:
            raise TypeError(f"Expected a bool, list[bool] or None for `nulls_first` but got {type(nulls_first)}")
        return MicroPartition._from_pymicropartition(self._micropartition.sort(pyexprs, descending, nulls_first))

    def sample(
        self,
        fraction: float | None = None,
        size: int | None = None,
        with_replacement: bool = False,
        seed: int | None = None,
    ) -> MicroPartition:
        if fraction is not None and size is not None:
            raise ValueError("Must specify either `fraction` or `size`, but not both")
        elif fraction is not None:
            return MicroPartition._from_pymicropartition(
                self._micropartition.sample_by_fraction(float(fraction), with_replacement, seed)
            )
        elif size is not None:
            return MicroPartition._from_pymicropartition(
                self._micropartition.sample_by_size(size, with_replacement, seed)
            )
        else:
            raise ValueError("Must specify either `fraction` or `size`")

    def agg(self, to_agg: list[Expression], group_by: ExpressionsProjection | None = None) -> MicroPartition:
        to_agg_pyexprs = [e._expr for e in to_agg]
        group_by_pyexprs = [e._expr for e in group_by] if group_by is not None else []
        return MicroPartition._from_pymicropartition(self._micropartition.agg(to_agg_pyexprs, group_by_pyexprs))

    def pivot(
        self, group_by: ExpressionsProjection, pivot_column: Expression, values_column: Expression, names: list[str]
    ) -> MicroPartition:
        group_by_pyexprs = [e._expr for e in group_by]
        pivot_column_pyexpr = pivot_column._expr
        values_column_pyexpr = values_column._expr
        return MicroPartition._from_pymicropartition(
            self._micropartition.pivot(group_by_pyexprs, pivot_column_pyexpr, values_column_pyexpr, names)
        )

    def quantiles(self, num: int) -> MicroPartition:
        return MicroPartition._from_pymicropartition(self._micropartition.quantiles(num))

    def explode(self, columns: ExpressionsProjection) -> MicroPartition:
        """NOTE: Expressions here must be Explode expressions."""
        to_explode_pyexprs = [e._expr for e in columns]
        return MicroPartition._from_pymicropartition(self._micropartition.explode(to_explode_pyexprs))

    def unpivot(
        self, ids: ExpressionsProjection, values: ExpressionsProjection, variable_name: str, value_name: str
    ) -> MicroPartition:
        ids_pyexprs = [e._expr for e in ids]
        values_pyexprs = [e._expr for e in values]
        return MicroPartition._from_pymicropartition(
            self._micropartition.unpivot(ids_pyexprs, values_pyexprs, variable_name, value_name)
        )

    def hash_join(
        self,
        right: MicroPartition,
        left_on: ExpressionsProjection,
        right_on: ExpressionsProjection,
        null_equals_nulls: list[bool] | None = None,
        how: JoinType = JoinType.Inner,
    ) -> MicroPartition:
        if len(left_on) != len(right_on):
            raise ValueError(
                f"Mismatch of number of join keys, left_on: {len(left_on)}, right_on: {len(right_on)}\nleft_on {left_on}\nright_on {right_on}"
            )

        if not isinstance(right, MicroPartition):
            raise TypeError(f"Expected a MicroPartition for `right` in join but got {type(right)}")

        left_exprs = [e._expr for e in left_on]
        right_exprs = [e._expr for e in right_on]

        return MicroPartition._from_pymicropartition(
            self._micropartition.hash_join(
                right._micropartition,
                left_on=left_exprs,
                right_on=right_exprs,
                null_equals_nulls=null_equals_nulls,
                how=how,
            )
        )

    def sort_merge_join(
        self,
        right: MicroPartition,
        left_on: ExpressionsProjection,
        right_on: ExpressionsProjection,
        how: JoinType = JoinType.Inner,
        is_sorted: bool = False,
    ) -> MicroPartition:
        if how != JoinType.Inner:
            raise NotImplementedError("TODO: [RUST] Implement Other Join types")
        if len(left_on) != len(right_on):
            raise ValueError(
                f"Mismatch of number of join keys, left_on: {len(left_on)}, right_on: {len(right_on)}\nleft_on {left_on}\nright_on {right_on}"
            )

        if not isinstance(right, MicroPartition):
            raise TypeError(f"Expected a MicroPartition for `right` in join but got {type(right)}")

        left_exprs = [e._expr for e in left_on]
        right_exprs = [e._expr for e in right_on]

        return MicroPartition._from_pymicropartition(
            self._micropartition.sort_merge_join(
                right._micropartition, left_on=left_exprs, right_on=right_exprs, is_sorted=is_sorted
            )
        )

    def cross_join(
        self,
        right: MicroPartition,
        outer_loop_side: JoinSide,
    ) -> MicroPartition:
        return MicroPartition._from_pymicropartition(
            self._micropartition.cross_join(right._micropartition, outer_loop_side=outer_loop_side)
        )

    def partition_by_hash(self, exprs: ExpressionsProjection, num_partitions: int) -> list[MicroPartition]:
        if not isinstance(num_partitions, int):
            raise TypeError(f"Expected a num_partitions to be int, got {type(num_partitions)}")

        pyexprs = [e._expr for e in exprs]
        return [
            MicroPartition._from_pymicropartition(t)
            for t in self._micropartition.partition_by_hash(pyexprs, num_partitions)
        ]

    def partition_by_range(
        self, partition_keys: ExpressionsProjection, boundaries: RecordBatch, descending: list[bool]
    ) -> list[MicroPartition]:
        if not isinstance(boundaries, RecordBatch):
            raise TypeError(f"Expected a RecordBatch for `boundaries` in partition_by_range but got {type(boundaries)}")

        exprs = [e._expr for e in partition_keys]
        return [
            MicroPartition._from_pymicropartition(t)
            for t in self._micropartition.partition_by_range(exprs, boundaries._recordbatch, descending)
        ]

    def partition_by_random(self, num_partitions: int, seed: int) -> list[MicroPartition]:
        if not isinstance(num_partitions, int):
            raise TypeError(f"Expected a num_partitions to be int, got {type(num_partitions)}")

        if not isinstance(seed, int):
            raise TypeError(f"Expected a seed to be int, got {type(seed)}")

        return [
            MicroPartition._from_pymicropartition(t)
            for t in self._micropartition.partition_by_random(num_partitions, seed)
        ]

    def partition_by_value(self, partition_keys: ExpressionsProjection) -> tuple[list[MicroPartition], MicroPartition]:
        exprs = [e._expr for e in partition_keys]
        PyRecordBatchs, values = self._micropartition.partition_by_value(exprs)
        return [
            MicroPartition._from_pymicropartition(t) for t in PyRecordBatchs
        ], MicroPartition._from_pymicropartition(values)

    def add_monotonically_increasing_id(self, partition_num: int, column_name: str) -> MicroPartition:
        return MicroPartition._from_pymicropartition(
            self._micropartition.add_monotonically_increasing_id(partition_num, column_name)
        )

    ###
    # Compute methods (MicroPartition -> Series)
    ###

    def argsort(
        self,
        sort_keys: ExpressionsProjection,
        descending: bool | list[bool] | None = None,
        nulls_first: bool | list[bool] | None = None,
    ) -> Series:
        assert all(isinstance(e, Expression) for e in sort_keys)
        pyexprs = [e._expr for e in sort_keys]
        if descending is None:
            descending = [False for _ in pyexprs]
        elif isinstance(descending, bool):
            descending = [descending for _ in pyexprs]
        elif isinstance(descending, list):
            if len(descending) != len(sort_keys):
                raise ValueError(
                    f"Expected length of `descending` to be the same length as `sort_keys` since a list was passed in,"
                    f"got {len(descending)} instead of {len(sort_keys)}"
                )
        else:
            raise TypeError(f"Expected a bool, list[bool] or None for `descending` but got {type(descending)}")
        if nulls_first is None:
            nulls_first = descending
        elif isinstance(nulls_first, bool):
            nulls_first = [nulls_first for _ in pyexprs]
        elif isinstance(nulls_first, list):
            if len(nulls_first) != len(sort_keys):
                raise ValueError(
                    f"Expected length of `nulls_first` to be the same length as `sort_keys` since a list was passed in,"
                    f"got {len(nulls_first)} instead of {len(sort_keys)}"
                )
            else:
                nulls_first = [bool(x) for x in nulls_first]
        else:
            raise TypeError(f"Expected a bool, list[bool] or None for `nulls_first` but got {type(nulls_first)}")
        return Series._from_pyseries(self._micropartition.argsort(pyexprs, descending, nulls_first))

    def __reduce__(self) -> tuple[Callable[[dict[str, Any]], MicroPartition], tuple[dict[str, Series]]]:
        names = self.column_names()
        return MicroPartition.from_pydict, ({name: self.get_column_by_name(name) for name in names},)

    @classmethod
    def read_parquet_statistics(
        cls,
        paths: Series | list[str],
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
    ) -> RecordBatch:
        return RecordBatch.read_parquet_statistics(paths=paths, io_config=io_config, multithreaded_io=multithreaded_io)

    @classmethod
    def read_parquet(
        cls,
        path: str,
        columns: list[str] | None = None,
        start_offset: int | None = None,
        num_rows: int | None = None,
        row_groups: list[int] | None = None,
        predicate: Expression | None = None,
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
        coerce_int96_timestamp_unit: TimeUnit = TimeUnit.ns(),
    ) -> MicroPartition:
        return MicroPartition._from_pymicropartition(
            _PyMicroPartition.read_parquet(
                path,
                columns,
                start_offset,
                num_rows,
                row_groups,
                predicate._expr if predicate is not None else None,
                io_config,
                multithreaded_io,
                coerce_int96_timestamp_unit._timeunit,
            )
        )

    @classmethod
    def read_parquet_bulk(
        cls,
        paths: list[str],
        columns: list[str] | None = None,
        start_offset: int | None = None,
        num_rows: int | None = None,
        row_groups_per_path: list[list[int] | None] | None = None,
        predicate: Expression | None = None,
        io_config: IOConfig | None = None,
        num_parallel_tasks: int | None = 128,
        multithreaded_io: bool | None = None,
        coerce_int96_timestamp_unit: TimeUnit = TimeUnit.ns(),
    ) -> MicroPartition:
        return MicroPartition._from_pymicropartition(
            _PyMicroPartition.read_parquet_bulk(
                paths,
                columns,
                start_offset,
                num_rows,
                row_groups_per_path,
                predicate._expr if predicate is not None else None,
                io_config,
                num_parallel_tasks,
                multithreaded_io,
                coerce_int96_timestamp_unit._timeunit,
            )
        )

    @classmethod
    def read_csv(
        cls,
        path: str,
        convert_options: CsvConvertOptions,
        parse_options: CsvParseOptions,
        read_options: CsvReadOptions,
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
    ) -> MicroPartition:
        return MicroPartition._from_pymicropartition(
            _PyMicroPartition.read_csv(
                uri=path,
                convert_options=convert_options,
                parse_options=parse_options,
                read_options=read_options,
                io_config=io_config,
                multithreaded_io=multithreaded_io,
            )
        )

    @classmethod
    def read_json(
        cls,
        path: str,
        convert_options: JsonConvertOptions | None = None,
        parse_options: JsonParseOptions | None = None,
        read_options: JsonReadOptions | None = None,
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
    ) -> MicroPartition:
        return MicroPartition._from_pymicropartition(
            _PyMicroPartition.read_json_native(
                uri=path,
                convert_options=convert_options,
                parse_options=parse_options,
                read_options=read_options,
                io_config=io_config,
                multithreaded_io=multithreaded_io,
            )
        )

    @classmethod
    def read_warc(
        cls,
        path: str,
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
    ) -> MicroPartition:
        return MicroPartition._from_pymicropartition(
            _PyMicroPartition.read_warc(
                uri=path,
                io_config=io_config,
                multithreaded_io=multithreaded_io,
            )
        )



from __future__ import annotations

import logging
import random
from collections import defaultdict, deque
from typing import TYPE_CHECKING

from daft.daft import (
    FlightClientManager,
    FlightServerConnectionHandle,
    InProgressShuffleCache,
    PyExpr,
    ShuffleCache,
    start_flight_server,
)
from daft.execution.execution_step import (
    PartitionTask,
    PartitionTaskBuilder,
    SingleOutputPartitionTask,
)
from daft.execution.physical_plan import InProgressPhysicalPlan, stage_id_counter
from daft.recordbatch.micropartition import MicroPartition
from daft.runners.partitioning import PartitionMetadata
from daft.runners.ray_runner import _ray_num_cpus_provider

if TYPE_CHECKING:
    from collections.abc import Generator

logger = logging.getLogger(__name__)

try:
    import ray
    import ray.experimental
    import ray.util.scheduling_strategies
    from ray._private.state import total_resources_per_node
except ImportError:
    logger.error(
        "Error when importing Ray. Please ensure that getdaft was installed with the Ray extras tag: getdaft[ray] (https://www.getdaft.io/projects/docs/en/latest/learn/install.html)"
    )
    raise


@ray.remote(num_cpus=0)
class ShuffleActorManager:
    """ShuffleActorManager manages the shuffle actors.

    It is responsible for:
    - Creating shuffle actors
    - Routing partitions to the correct actor based on the object location
    - Waiting for all the partitions to be pushed to the actors
    - Sending clear partition requests to the actors
    - Shutting down the actors
    """

    def __init__(
        self,
        shuffle_stage_id: int,
        storage_dirs: list[str],
        num_output_partitions: int,
        partition_by: list[PyExpr] | None = None,
    ):
        self.shuffle_stage_id = shuffle_stage_id
        self.storage_dirs = storage_dirs
        self.num_output_partitions = num_output_partitions
        self.partition_by = partition_by

        self.all_actors: dict[str, ray.actor.ActorHandle] = {}
        self.active_actors: dict[str, ray.actor.ActorHandle] = {}

        self.fetch_partition_futures: list[ray.ObjectRef] = []
        self.clear_partition_futures: list[ray.ObjectRef] = []

        # Eagerly create actors for all nodes that are currently available (with cpu)
        for node in ray.nodes():
            if "Resources" in node and "CPU" in node["Resources"] and node["Resources"]["CPU"] > 0:
                self.get_or_create_actor(node["NodeID"])

    # Get or create an actor for a given node id
    def get_or_create_actor(self, node_id: str) -> ray.actor.ActorHandle:
        if node_id not in self.all_actors:
            num_cpus = total_resources_per_node()[node_id]["CPU"]
            actor = ShuffleActor.options(  # type: ignore[attr-defined]
                scheduling_strategy=ray.util.scheduling_strategies.NodeAffinitySchedulingStrategy(
                    node_id=node_id,
                    soft=False,  # TODO: check if this is dangerous, if so, check if we can do true
                ),
                max_concurrency=int(num_cpus * 2),
            ).remote(
                ray.get_runtime_context().current_actor,
                self.shuffle_stage_id,
                self.storage_dirs,
                self.num_output_partitions,
                self.partition_by,
            )
            self.all_actors[node_id] = actor
        return self.all_actors[node_id]

    def get_active_actors(self) -> list[ray.actor.ActorHandle]:
        return list(self.active_actors.values())

    def get_active_node_ids(self) -> list[str]:
        return list(self.active_actors.keys())

    def get_active_actor_addresses(self) -> list[str]:
        return ray.get([self.active_actors[node_id].get_address.remote() for node_id in self.active_actors])

    def get_all_actors(self) -> list[ray.actor.ActorHandle]:
        return list(self.all_actors.values())

    def get_all_node_ids(self) -> list[str]:
        return list(self.all_actors.keys())

    def get_all_actor_addresses(self) -> list[str]:
        return ray.get([self.all_actors[node_id].get_address.remote() for node_id in self.all_actors])

    # Push an object to an actor
    def _push_objects_to_actor(self, node_id: str, objects: list[ray.ObjectRef]) -> ray.ObjectRef:
        actor = self.get_or_create_actor(node_id)
        self.active_actors[node_id] = actor
        return actor.push_partitions.remote(*objects)

    # Push a list of objects to the actors
    def push_objects(self, objects: list[ray.ObjectRef]) -> None:
        object_locations = ray.experimental.get_object_locations(objects)
        assert len(object_locations) == len(objects)

        objects_to_push_per_node = defaultdict(list)
        for object, object_location in object_locations.items():
            if object_location is None or "node_ids" not in object_location or not object_location["node_ids"]:
                # if we don't know where the object is, push it to a random actor
                node_id = random.choice(list(self.all_actors.keys()))
                objects_to_push_per_node[node_id].append(object)
            else:
                # there might be multiple nodes that have the object, so we push it to a random one.
                if len(object_location["node_ids"]) > 1:
                    node_id = random.choice(object_location["node_ids"])
                else:
                    [node_id] = object_location["node_ids"]
                objects_to_push_per_node[node_id].append(object)

        ray.get(
            [self._push_objects_to_actor(node_id, objects) for node_id, objects in objects_to_push_per_node.items()]
        )

    # Wait for all shuffle actors to finish pushing objects to the cache
    def finish_push_objects(self) -> list[PartitionMetadata]:
        metadatas_per_actor = []
        for actor in self.active_actors.values():
            metadatas_per_actor.append(
                actor.finish_push_partitions.options(num_returns=self.num_output_partitions).remote()
            )

        metadatas_per_actor = [ray.get(metadatas) for metadatas in metadatas_per_actor]

        merged_metadatas = []
        for metadatas in zip(*metadatas_per_actor):
            rows = sum([metadata.num_rows for metadata in metadatas])
            bytes = sum([metadata.size_bytes for metadata in metadatas])
            merged_metadatas.append(PartitionMetadata(num_rows=rows, size_bytes=bytes))
        return merged_metadatas

    # Clear the given partitions from the shuffle actors
    def clear_partition(self, partition_idx: int) -> None:
        self.clear_partition_futures.extend(
            self.active_actors[node_id].clear_partition.remote(partition_idx) for node_id in self.active_actors
        )

    # Shutdown the shuffle actor manager and all the shuffle actors
    def shutdown(self) -> None:
        ray.get(self.clear_partition_futures)
        self.clear_partition_futures.clear()

        ray.get([actor.shutdown.remote() for actor in self.all_actors.values()])
        for actor in self.all_actors.values():
            ray.kill(actor)
        self.all_actors.clear()


@ray.remote(num_cpus=0)
class ShuffleActor:
    """ShuffleActor is a per-node actor that manages the map partitions produced on that node.

    It is responsible for:
    - Partitioning the input data into num_partitions
    - Saving the partitions to a cache (for now, it defaults write to disk)
    - Serving a Flight server to fetch the partitions for the reduce phase
    - Cleaning up the cache when the reduce phase is done
    - Shutting down the Flight server
    """

    def __init__(
        self,
        shuffle_actor_manager: ray.actor.ActorHandle,
        shuffle_stage_id: int,
        storage_dirs: list[str],
        num_output_partitions: int,
        partition_by: list[PyExpr] | None = None,
    ):
        self.shuffle_actor_manager = shuffle_actor_manager
        self.node_id = ray.get_runtime_context().get_node_id()
        self.host = ray.util.get_node_ip_address()

        self.num_output_partitions = num_output_partitions
        self.partition_by = partition_by

        # create a shuffle cache to store the partitions
        self.in_progress_shuffle_cache: InProgressShuffleCache | None = InProgressShuffleCache.try_new(
            num_output_partitions,
            storage_dirs,
            self.node_id,
            shuffle_stage_id,
            target_filesize=1024 * 1024 * 10,
            compression=None,
            partition_by=partition_by,
        )
        self.shuffle_cache: ShuffleCache | None = None
        self.client_manager: FlightClientManager | None = None

        self.server: FlightServerConnectionHandle | None = None
        self.port: int | None = None

    def get_address(self) -> str:
        return f"grpc://{self.host}:{self.port}"

    # Push a partition to the shuffle cache asynchronously
    async def push_partitions(self, *partitions: MicroPartition) -> None:
        assert self.in_progress_shuffle_cache is not None, "Shuffle cache not initialized"
        await self.in_progress_shuffle_cache.push_partitions([partition._micropartition for partition in partitions])

    # Wait for all the push partition futures to complete
    async def finish_push_partitions(self) -> list[PartitionMetadata]:
        assert self.in_progress_shuffle_cache is not None, "Shuffle cache not initialized"
        self.shuffle_cache = await self.in_progress_shuffle_cache.close()
        self.server = start_flight_server(self.shuffle_cache, self.host)
        self.port = self.server.port()
        self.in_progress_shuffle_cache = None
        rows_per_partition = self.shuffle_cache.rows_per_partition()
        bytes_per_partition = self.shuffle_cache.bytes_per_partition()
        metadatas = [
            PartitionMetadata(num_rows=rows, size_bytes=bytes)
            for rows, bytes in zip(rows_per_partition, bytes_per_partition)
        ]
        return metadatas

    def initialize_client_manager(self, addresses: list[str], num_parallel_fetches: int = 2) -> None:
        assert self.client_manager is None, "Client manager already initialized"
        assert self.shuffle_cache is not None, "Shuffle cache not initialized"
        schema = self.shuffle_cache.schema()
        self.client_manager = FlightClientManager(addresses, num_parallel_fetches, schema)

    async def fetch_partition(self, partition_idx: int) -> tuple[MicroPartition, ray.ObjectRef]:
        assert self.client_manager is not None, "Client manager not initialized"
        py_mp = await self.client_manager.fetch_partition(partition_idx)
        clear_partition_future = self.shuffle_actor_manager.clear_partition.remote(partition_idx)
        return MicroPartition._from_pymicropartition(py_mp), clear_partition_future

    # Clean up the shuffle files for the given partition
    def clear_partition(self, partition_idx: int) -> None:
        if self.shuffle_cache is not None:
            try:
                self.shuffle_cache.clear_partition(partition_idx)
            except Exception as e:
                print(f"failed to clean up shuffle files: {e} for partition {partition_idx}")

    # Shutdown the shuffle actor
    def shutdown(self) -> None:
        if self.server is not None:
            self.server.shutdown()
        if self.shuffle_cache is not None:
            try:
                self.shuffle_cache.clear_directories()
            except Exception as e:
                print(f"failed to clean up shuffle files: {e}")


def run_map_phase(
    plan: InProgressPhysicalPlan[ray.ObjectRef],
    map_stage_id: int,
    shuffle_actor_manager: ray.actor.ActorHandle,
) -> Generator[
    None | PartitionTask[ray.ObjectRef] | PartitionTaskBuilder[ray.ObjectRef], None, list[PartitionMetadata]
]:
    # Maps tasks we have not emitted
    pending_map_tasks: deque[SingleOutputPartitionTask[ray.ObjectRef]] = deque()
    # Maps tasks we have emitted but not yet completed
    in_flight_map_tasks: dict[str, SingleOutputPartitionTask[ray.ObjectRef]] = {}

    # Collect all the map tasks
    for step in plan:
        if isinstance(step, PartitionTaskBuilder):
            step = step.finalize_partition_task_single_output(map_stage_id)
            pending_map_tasks.append(step)
        else:
            yield step

    while pending_map_tasks or in_flight_map_tasks:
        # Get all the map tasks that are done
        done_ids = [id for id in in_flight_map_tasks.keys() if in_flight_map_tasks[id].done()]

        # If there are any map tasks that are done, push them to the actor manager
        if len(done_ids) > 0:
            done_partitions = [in_flight_map_tasks.pop(id).partition() for id in done_ids]
            ray.get(shuffle_actor_manager.push_objects.remote(done_partitions))
        # If there are no map tasks that are done, try to emit more map tasks. Else, yield None to indicate to the scheduler that we need to wait
        else:
            # Max number of maps to emit per wave
            batch_size = next(_ray_num_cpus_provider())
            available_to_yield = min(batch_size - len(in_flight_map_tasks), len(pending_map_tasks))
            if available_to_yield > 0:
                for _ in range(available_to_yield):
                    map_task = pending_map_tasks.popleft()
                    in_flight_map_tasks[map_task.id()] = map_task
                    yield map_task
            else:
                yield None

    # Wait for all actors to complete the map phase
    metadatas = ray.get(shuffle_actor_manager.finish_push_objects.remote())
    return metadatas


def run_reduce_phase(
    num_output_partitions: int,
    shuffle_actor_manager: ray.actor.ActorHandle,
    metadatas: list[PartitionMetadata],
) -> Generator[PartitionTaskBuilder[ray.ObjectRef], None, None]:
    # only the active actors have data from the map
    active_actor_addresses = ray.get(shuffle_actor_manager.get_active_actor_addresses.remote())
    # but we can do reduce on all actors
    all_actors = ray.get(shuffle_actor_manager.get_all_actors.remote())

    # initialize the client manager for all actors
    ray.get([actor.initialize_client_manager.remote(active_actor_addresses) for actor in all_actors])

    clear_partition_futures = []
    for partition_idx in range(num_output_partitions):
        partition_ref, clear_partition_future = (
            all_actors[partition_idx % len(all_actors)].fetch_partition.options(num_returns=2).remote(partition_idx)
        )
        clear_partition_futures.append(clear_partition_future)
        yield PartitionTaskBuilder(
            inputs=[partition_ref],
            partial_metadatas=[metadatas[partition_idx]],
        )

    # we need to wait for all the clear partition futures to complete before shutting down the actor manager
    ray.get(clear_partition_futures)
    ray.get(shuffle_actor_manager.shutdown.remote())
    ray.kill(shuffle_actor_manager)


def flight_shuffle(
    fanout_plan: InProgressPhysicalPlan[ray.ObjectRef],
    num_output_partitions: int,
    shuffle_dirs: list[str],
    partition_by: list[PyExpr] | None = None,
) -> Generator[None | PartitionTask[ray.ObjectRef] | PartitionTaskBuilder[ray.ObjectRef], None, None]:
    map_stage_id = next(stage_id_counter)
    shuffle_stage_id = next(stage_id_counter)
    # Try to schedule the manager on the current node, which should be the head node
    shuffle_actor_manager = ShuffleActorManager.options(  # type: ignore[attr-defined]
        scheduling_strategy=ray.util.scheduling_strategies.NodeAffinitySchedulingStrategy(
            node_id=ray.get_runtime_context().get_node_id(),
            soft=False,
        )
    ).remote(
        shuffle_stage_id,
        shuffle_dirs,
        num_output_partitions,
        partition_by,
    )

    # Run the map phase
    metadatas = yield from run_map_phase(
        fanout_plan,
        map_stage_id,
        shuffle_actor_manager,
    )

    # Run the reduce phase
    yield from run_reduce_phase(num_output_partitions, shuffle_actor_manager, metadatas)



from __future__ import annotations

import random

import pandas as pd
import pytest

from daft import Window, col
from daft.functions import dense_rank, rank, row_number
from tests.conftest import assert_df_equals, get_tests_daft_runner_name

pytestmark = pytest.mark.skipif(
    get_tests_daft_runner_name() != "native", reason="Window tests only run on native runner"
)


@pytest.mark.parametrize("repartition_nparts", [1, 2, 20, 50, 100])
def test_row_number(make_df, repartition_nparts, with_morsel_size):
    df = make_df(
        {"category": ["A", "A", "A", "B", "B", "B", "C", "C"], "sales": [100, 200, 50, 500, 125, 300, 250, 150]},
        repartition=repartition_nparts,
        repartition_columns=["category"],
    )

    window_spec = Window().order_by("sales", desc=False)

    result = df.select(
        col("category"), col("sales"), row_number().over(window_spec).alias("row_number_sales")
    ).collect()

    expected = {
        "category": ["A", "A", "A", "B", "B", "B", "C", "C"],
        "sales": [100, 200, 50, 500, 125, 300, 250, 150],
        "row_number_sales": [2, 5, 1, 8, 3, 7, 6, 4],
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=list(expected.keys()), check_dtype=False)


@pytest.mark.parametrize("repartition_nparts", [1, 2, 20, 50, 100])
def test_order_by_only_row_number(make_df, repartition_nparts, with_morsel_size):
    """Test row_number function with order_by only (no partition_by)."""
    random.seed(42)

    data = []
    n = 1000
    total = n + 1

    xs = list(range(1, n + 1))
    ys = list(range(1, n + 1))
    values = list(range(1, n + 1))

    random.shuffle(xs)
    random.shuffle(ys)
    random.shuffle(values)

    for x, y, value in zip(xs, ys, values):
        data.append({"x": x, "y": y, "value": value})

    df = make_df(data, repartition=repartition_nparts, repartition_columns=["x", "y"])

    window_spec_x = Window().order_by("x", desc=False)
    window_spec_y = Window().order_by("y", desc=False)
    window_spec_value = Window().order_by("value", desc=False)

    window_spec_x_desc = Window().order_by("x", desc=True)
    window_spec_y_desc = Window().order_by("y", desc=True)
    window_spec_value_desc = Window().order_by("value", desc=True)

    window_spec_xy = Window().order_by(["x", "y"], desc=[False, False])
    window_spec_xy_mixed = Window().order_by(["x", "y"], desc=[False, True])

    result = df.select(
        col("x"),
        col("y"),
        col("value"),
        row_number().over(window_spec_x).alias("row_by_x_asc"),
        row_number().over(window_spec_y).alias("row_by_y_asc"),
        row_number().over(window_spec_value).alias("row_by_value_asc"),
        row_number().over(window_spec_x_desc).alias("row_by_x_desc"),
        row_number().over(window_spec_y_desc).alias("row_by_y_desc"),
        row_number().over(window_spec_value_desc).alias("row_by_value_desc"),
        row_number().over(window_spec_xy).alias("row_by_xy_asc"),
        row_number().over(window_spec_xy_mixed).alias("row_by_xy_mixed"),
    ).collect()

    result_df = result.to_pandas()

    for _, row in result_df.iterrows():
        assert row["row_by_x_asc"] == row["x"], f"row_by_x_asc {row['row_by_x_asc']} should equal x {row['x']}"
        assert row["row_by_y_asc"] == row["y"], f"row_by_y_asc {row['row_by_y_asc']} should equal y {row['y']}"
        assert (
            row["row_by_value_asc"] == row["value"]
        ), f"row_by_value_asc {row['row_by_value_asc']} should equal value {row['value']}"

        assert (
            row["row_by_x_desc"] == total - row["x"]
        ), f"row_by_x_desc {row['row_by_x_desc']} should equal {total} - x {row['x']}"
        assert (
            row["row_by_y_desc"] == total - row["y"]
        ), f"row_by_y_desc {row['row_by_y_desc']} should equal {total} - y {row['y']}"
        assert (
            row["row_by_value_desc"] == total - row["value"]
        ), f"row_by_value_desc {row['row_by_value_desc']} should equal {total} - value {row['value']}"

        assert row["row_by_xy_asc"] == row["x"], f"row_by_xy_asc {row['row_by_xy_asc']} should equal x {row['x']}"
        assert row["row_by_xy_mixed"] == row["x"], f"row_by_xy_mixed {row['row_by_xy_mixed']} should equal x {row['x']}"


@pytest.mark.parametrize("repartition_nparts", [1, 2, 20, 50, 100])
def test_rank(make_df, repartition_nparts, with_morsel_size):
    """Test rank function with order_by only (no partition_by)."""
    random.seed(42)

    data = []
    n = 1000

    values = []
    for i in range(1, 11):
        values.extend([i] * 100)

    values = values[:n]

    random.shuffle(values)
    categories = ["A", "B", "C", "D", "E"]

    for i, value in enumerate(values):
        data.append({"id": i, "category": random.choice(categories), "value": value})

    df = make_df(data, repartition=repartition_nparts, repartition_columns=["id", "category"])

    window_spec_asc = Window().order_by("value", desc=False)
    window_spec_desc = Window().order_by("value", desc=True)

    result = df.select(
        col("id"),
        col("category"),
        col("value"),
        rank().over(window_spec_asc).alias("rank_asc"),
        rank().over(window_spec_desc).alias("rank_desc"),
    ).collect()

    result_df = result.to_pandas()

    value_to_rank_asc = {}
    current_rank = 1
    for value in sorted(values):
        if value not in value_to_rank_asc:
            value_to_rank_asc[value] = current_rank
        current_rank += 1

    value_to_rank_desc = {}
    current_rank = 1
    for value in sorted(values, reverse=True):
        if value not in value_to_rank_desc:
            value_to_rank_desc[value] = current_rank
        current_rank += 1

    expected_rank_asc = []
    expected_rank_desc = []
    for value in values:
        expected_rank_asc.append(value_to_rank_asc[value])
        expected_rank_desc.append(value_to_rank_desc[value])

    expected = pd.DataFrame(
        {
            "id": list(range(n)),
            "category": [data[i]["category"] for i in range(n)],
            "value": values,
            "rank_asc": expected_rank_asc,
            "rank_desc": expected_rank_desc,
        }
    )

    result_df = result_df.sort_values(by="id").reset_index(drop=True)
    expected = expected.sort_values(by="id").reset_index(drop=True)

    assert_df_equals(result_df, expected, sort_key=["id", "category", "value"], check_dtype=False)


@pytest.mark.parametrize("repartition_nparts", [1, 2, 20, 50, 100])
def test_dense_rank(make_df, repartition_nparts, with_morsel_size):
    """Test dense_rank function with order_by only (no partition_by)."""
    random.seed(43)

    data = []
    n = 1000

    values = []
    for i in range(1, 11):
        values.extend([i] * 100)

    values = values[:n]

    random.shuffle(values)

    categories = ["A", "B", "C", "D", "E"]

    for i, value in enumerate(values):
        data.append({"id": i, "category": random.choice(categories), "value": value})

    df = make_df(data, repartition=repartition_nparts, repartition_columns=["id", "category"])

    window_spec_asc = Window().order_by("value", desc=False)
    window_spec_desc = Window().order_by("value", desc=True)

    result = df.select(
        col("id"),
        col("category"),
        col("value"),
        dense_rank().over(window_spec_asc).alias("dense_rank_asc"),
        dense_rank().over(window_spec_desc).alias("dense_rank_desc"),
    ).collect()

    result_df = result.to_pandas()

    unique_values_asc = sorted(set(values))
    unique_values_desc = sorted(set(values), reverse=True)

    value_to_dense_rank_asc = {value: i + 1 for i, value in enumerate(unique_values_asc)}
    value_to_dense_rank_desc = {value: i + 1 for i, value in enumerate(unique_values_desc)}

    expected = pd.DataFrame(
        {
            "id": list(range(n)),
            "category": [data[i]["category"] for i in range(n)],
            "value": values,
            "dense_rank_asc": [value_to_dense_rank_asc[value] for value in values],
            "dense_rank_desc": [value_to_dense_rank_desc[value] for value in values],
        }
    )

    result_df = result_df.sort_values(by="id").reset_index(drop=True)
    expected = expected.sort_values(by="id").reset_index(drop=True)

    assert_df_equals(result_df, expected, sort_key=["id", "category", "value"], check_dtype=False)


@pytest.mark.parametrize("repartition_nparts", [1, 2, 20, 50, 100])
def test_rank_with_1k_distinct_values(make_df, repartition_nparts, with_morsel_size):
    """Test rank, dense_rank, and row_number with 15,000 distinct values."""
    n = 1000
    random.seed(45)

    values = list(range(n))
    random.shuffle(values)

    data = [{"id": i, "value": value} for i, value in enumerate(values)]

    df = make_df(data, repartition=repartition_nparts, repartition_columns=["id"])

    window_asc = Window().order_by("value", desc=False)
    window_desc = Window().order_by("value", desc=True)

    result = df.select(
        col("id"),
        col("value"),
        row_number().over(window_asc).alias("row_number_asc"),
        rank().over(window_asc).alias("rank_asc"),
        dense_rank().over(window_asc).alias("dense_rank_asc"),
        row_number().over(window_desc).alias("row_number_desc"),
        rank().over(window_desc).alias("rank_desc"),
        dense_rank().over(window_desc).alias("dense_rank_desc"),
    ).collect()

    result_df = result.to_pandas().sort_values("id").reset_index(drop=True)

    value_to_row = {v: i + 1 for i, v in enumerate(sorted(values))}
    value_to_rev_row = {v: i + 1 for i, v in enumerate(sorted(values, reverse=True))}

    expected_df = (
        pd.DataFrame(
            {
                "id": list(range(n)),
                "value": values,
                "row_number_asc": [value_to_row[v] for v in values],
                "rank_asc": [value_to_row[v] for v in values],
                "dense_rank_asc": [value_to_row[v] for v in values],
                "row_number_desc": [value_to_rev_row[v] for v in values],
                "rank_desc": [value_to_rev_row[v] for v in values],
                "dense_rank_desc": [value_to_rev_row[v] for v in values],
            }
        )
        .sort_values("id")
        .reset_index(drop=True)
    )

    assert_df_equals(result_df, expected_df, sort_key=["id", "value"], check_dtype=False)



from __future__ import annotations

import dataclasses
import warnings
from typing import TYPE_CHECKING, Callable, Literal
from urllib.parse import urlparse

import unitycatalog

from daft.io import AzureConfig, IOConfig, S3Config

if TYPE_CHECKING:
    from unitycatalog.types import TableInfo


@dataclasses.dataclass(frozen=True)
class UnityCatalogTable:
    table_info: TableInfo
    table_uri: str
    io_config: IOConfig | None


class UnityCatalog:
    """Client to access the Unity Catalog.

    Unity Catalog is an open-sourced data catalog that can be self-hosted, or hosted by Databricks.

    Example of reading a dataframe from a table in Unity Catalog hosted by Databricks:

    >>> cat = UnityCatalog("https://<databricks_workspace_id>.cloud.databricks.com", token="my-token")
    >>> table = cat.load_table("my_catalog.my_schema.my_table")
    >>> df = daft.read_deltalake(table)
    """

    def __init__(self, endpoint: str, token: str | None = None):
        self._client = unitycatalog.Unitycatalog(
            base_url=endpoint.rstrip("/") + "/api/2.1/unity-catalog/",
            default_headers={"Authorization": f"Bearer {token}"},
        )

    def _paginate_to_completion(
        self,
        client_func_call: Callable[[unitycatalog.Unitycatalog, str | None], tuple[list[str] | None, str | None]],
    ) -> list[str]:
        results = []

        # Make first request
        new_results, next_page_token = client_func_call(self._client, None)
        if new_results is not None:
            results.extend(new_results)

        # Exhaust pages
        while next_page_token is not None and next_page_token != "":
            new_results, next_page_token = client_func_call(self._client, next_page_token)
            if new_results is not None:
                results.extend(new_results)

        return results

    def list_catalogs(self) -> list[str]:
        def _paginated_list_catalogs(
            client: unitycatalog.Unitycatalog, page_token: str | None
        ) -> tuple[list[str] | None, str | None]:
            response = client.catalogs.list(page_token=page_token)
            next_page_token = response.next_page_token
            if response.catalogs is None:
                return None, next_page_token
            return [c.name for c in response.catalogs], next_page_token

        return self._paginate_to_completion(_paginated_list_catalogs)

    def list_schemas(self, catalog_name: str) -> list[str]:
        def _paginated_list_schemas(
            client: unitycatalog.Unitycatalog, page_token: str | None
        ) -> tuple[list[str] | None, str | None]:
            response = client.schemas.list(catalog_name=catalog_name, page_token=page_token)
            next_page_token = response.next_page_token
            if response.schemas is None:
                return None, next_page_token
            return [s.full_name for s in response.schemas], next_page_token

        return self._paginate_to_completion(_paginated_list_schemas)

    def list_tables(self, schema_name: str) -> list[str]:
        if schema_name.count(".") != 1:
            raise ValueError(
                f"Expected fully-qualified schema name with format `catalog_name`.`schema_name`, but received: {schema_name}"
            )

        catalog_name, schema_name = schema_name.split(".")

        def _paginated_list_tables(
            client: unitycatalog.Unitycatalog, page_token: str | None
        ) -> tuple[list[str] | None, str | None]:
            response = client.tables.list(catalog_name=catalog_name, schema_name=schema_name, page_token=page_token)
            next_page_token = response.next_page_token
            if response.tables is None:
                return None, next_page_token
            return [f"{t.catalog_name}.{t.schema_name}.{t.name}" for t in response.tables], next_page_token

        return self._paginate_to_completion(_paginated_list_tables)

    def load_table(
        self,
        table_name: str,
        new_table_storage_path: str | None = None,
        operation: Literal["READ", "READ_WRITE"] = "READ_WRITE",
        table_type: Literal["EXTERNAL", "MANAGED"] = "EXTERNAL",
    ) -> UnityCatalogTable:
        """Loads an existing Unity Catalog table. If the table is not found, and information is provided in the method to create a new table, a new table will be attempted to be registered.

        Args:
            table_name (str): Name of the table in Unity Catalog in the form of dot-separated, 3-level namespace
            new_table_storage_path (str, optional): Cloud storage path URI to register a new external table using this path. Unity Catalog will validate if the path is valid and authorized for the principal, else will raise an exception.
            operation ("READ" or "READ_WRITE", optional): The intended use of the table, which impacts authorization. Defaults to "READ_WRITE".
                table_type ("EXTERNAL" or "MANAGED", optional): The table type impacts the creation of the table if it does not exist. If "MANAGED" the table is created as managed table. If "EXTERNAL" the table is created based on the new_tabke_storage_path.

        Returns:
            UnityCatalogTable
        """
        # Load the table ID
        try:
            table_info = self._client.tables.retrieve(table_name)
            if new_table_storage_path:
                warnings.warn(
                    f"Table {table_name} is an existing storage table with a valid storage path. The 'new_table_storage_path' argument provided will be ignored."
                )
        except unitycatalog.NotFoundError:
            if table_type == "EXTERNAL" and not new_table_storage_path:
                raise ValueError(
                    f"Table {table_name} is not an existing table. If a new table needs to be created, provide 'new_table_storage_path' value or set table_type to 'MANAGED'."
                )
            elif table_type == "MANAGED":
                warnings.warn(f"Table {table_name} is being created as managed table in Unity Catalog.")
            try:
                three_part_namesplit = table_name.split(".")
                if len(three_part_namesplit) != 3 or not all(three_part_namesplit):
                    raise ValueError(
                        f"Expected table name to be in the format of 'catalog.schema.table', received: {table_name}"
                    )

                params = {
                    "catalog_name": three_part_namesplit[0],
                    "schema_name": three_part_namesplit[1],
                    "name": three_part_namesplit[2],
                    "columns": None,
                    "data_source_format": "DELTA",
                    "table_type": table_type,
                    "storage_location": new_table_storage_path,
                    "comment": None,
                }

                table_info = self._client.tables.create(**params)
            except Exception as e:
                raise Exception(f"An error occurred while registering the table in Unity Catalog: {e}")

        table_id = table_info.table_id
        storage_location = table_info.storage_location
        # Grab credentials from Unity catalog and place it into the Table
        temp_table_credentials = self._client.temporary_table_credentials.create(operation=operation, table_id=table_id)

        scheme = urlparse(storage_location).scheme
        if scheme == "s3" or scheme == "s3a":
            aws_temp_credentials = temp_table_credentials.aws_temp_credentials
            io_config = (
                IOConfig(
                    s3=S3Config(
                        key_id=aws_temp_credentials.access_key_id,
                        access_key=aws_temp_credentials.secret_access_key,
                        session_token=aws_temp_credentials.session_token,
                    )
                )
                if aws_temp_credentials is not None
                else None
            )
        elif scheme == "gcs" or scheme == "gs":
            # TO-DO: gather GCS credential vending assets from Unity and construct 'io_config``
            warnings.warn("GCS credential vending from Unity Catalog is not yet supported.")
            io_config = None
        elif scheme == "az" or scheme == "abfs" or scheme == "abfss":
            io_config = IOConfig(
                azure=AzureConfig(sas_token=temp_table_credentials.azure_user_delegation_sas.get("sas_token"))
            )
        else:
            warnings.warn(f"Credentials for scheme {scheme} are not yet supported.")
            io_config = None

        return UnityCatalogTable(
            table_info=table_info,
            table_uri=storage_location,
            io_config=io_config,
        )



from __future__ import annotations

import pytest

import daft
from daft import col

# see:
#  - https://docs.python.org/3/library/codecs.html#text-encodings
#  - https://docs.python.org/3/library/codecs.html#binary-transforms

TEXT = b"""
Every great magic trick consists of three parts or acts. The first part is
called "The Pledge". The magician shows you something ordinary: a deck of
cards, a bird or a man. He shows you this object. Perhaps he asks you to
inspect it to see if it is indeed real, unaltered, normal. But of course...
it probably isn't. The second act is called "The Turn". The magician takes
the ordinary something and makes it do something extraordinary. Now you're
looking for the secret... but you won't find it, because of course you're
not really looking. You don't really want to know. You want to be fooled.
But you wouldn't clap yet. Because making something disappear isn't enough;
you have to bring it back. That's why every magic trick has a third act,
the hardest part, the part we call "The Prestige". - Christopher Priest.
""" + bytes("🔥", encoding="utf-8")

###
# helpers
###


def _assert_eq(actual, expect):
    actual.to_pydict() == expect.to_pydict()


def _test_codec(codec: str, buff: bytes):
    _test_encode(codec, input=TEXT, output=buff)
    _test_decode(codec, input=buff, output=TEXT)
    _test_roundtrip(codec, input=TEXT)


def _test_encode(codec: str, input: bytes, output: bytes):
    df = daft.from_pydict({"v": [input]})
    actual = df.select(col("v").encode(codec))
    expect = daft.from_pydict({"v": [output]})
    _assert_eq(actual, expect)


def _test_decode(codec: str, input: bytes, output: bytes):
    df = daft.from_pydict({"v": [input]})
    actual = df.select(col("v").decode(codec))
    expect = daft.from_pydict({"v": [output]})
    _assert_eq(actual, expect)


def _test_roundtrip(codec: str, input: bytes):
    df1 = daft.from_pydict({"v": [input]})
    df2 = df1.select(col("v").encode(codec).decode(codec))
    _assert_eq(df1, df2)


def test_null_handling():
    import zlib

    #
    df1 = daft.from_pydict({"v": [None, zlib.compress(TEXT)]})
    df2 = daft.from_pydict({"v": [zlib.compress(TEXT), None]})
    result1 = df1.select(col("v").decode("zlib")).to_pydict()
    result2 = df2.select(col("v").decode("zlib")).to_pydict()
    #
    # assert validity is included
    assert result1["v"][0] is None
    assert result2["v"][1] is None


def test_with_strings():
    import zlib

    _test_encode("zlib", input=str(TEXT), output=zlib.compress(TEXT))


###
# codecs
###


def test_codec_deflate():
    import zlib

    # https://stackoverflow.com/questions/1089662/python-inflate-and-deflate-implementations
    # strip 2-byte zlib header and the 4-byte checksum
    _test_codec("deflate", buff=zlib.compress(TEXT)[2:-4])


def test_codec_gzip():
    import gzip

    _test_codec("gz", buff=gzip.compress(TEXT))
    _test_codec("gzip", buff=gzip.compress(TEXT))


def test_codec_zlib():
    import zlib

    _test_codec("zlib", buff=zlib.compress(TEXT))


def test_codec_base64():
    with pytest.raises(Exception, match="unsupported codec"):
        _test_codec("base64", None)


def test_codec_zstd():
    with pytest.raises(Exception, match="unsupported codec"):
        _test_codec("zstd", None)


def test_codec_bz2():
    with pytest.raises(Exception, match="unsupported codec"):
        _test_codec("bz2", None)


###
# utf-8 special handling
###


def test_decode_utf8():
    df = daft.from_pydict({"bytes": [None, TEXT]})
    actual = df.select(col("bytes").decode("utf-8"))
    expect = {"bytes": [None, TEXT.decode("utf-8")]}
    assert actual.to_pydict() == expect


def test_try_decode_utf8():
    # source
    # https://stackoverflow.com/questions/1301402/example-invalid-utf8-string
    # https://www.php.net/manual/en/reference.pcre.pattern.modifiers.php#54805
    valid_ascii = b"a"
    valid_2_octet = b"\xc3\xb1"
    invalid_2_octet = b"\xc3\x28"
    invalid_seq_id = b"\xa0\xa1"
    valid_3_octet = b"\xe2\x82\xa1"
    invalid_3_octet_2nd = b"\xe2\x28\xa1"
    invalid_3_octet_3rd = b"\xe2\x82\x28"
    valid_4_octet = b"\xf0\x90\x8c\xbc"
    invalid_4_octet_2nd = b"\xf0\x28\x8c\xbc"
    invalid_4_octet_3rd = b"\xf0\x90\x28\xbc"
    invalid_4_octet_4th = b"\xf0\x28\x8c\x28"
    valid_5_octet = b"\xf8\xa1\xa1\xa1\xa1"
    valid_6_octet = b"\xfc\xa1\xa1\xa1\xa1\xa1"

    df = daft.from_pydict(
        {
            "bytes": [
                None,
                TEXT,
                valid_ascii,
                valid_2_octet,
                invalid_2_octet,
                invalid_seq_id,
                valid_3_octet,
                invalid_3_octet_2nd,
                invalid_3_octet_3rd,
                valid_4_octet,
                invalid_4_octet_2nd,
                invalid_4_octet_3rd,
                invalid_4_octet_4th,
                valid_5_octet,
                valid_6_octet,
            ],
            "note": [
                "None",
                "Valid TEXT",
                "Valid ASCII",
                "Valid 2 Octet Sequence",
                "Invalid 2 Octet Sequence",
                "Invalid Sequence Identifier",
                "Valid 3 Octet Sequence",
                "Invalid 3 Octet Sequence (in 2nd Octet)",
                "Invalid 3 Octet Sequence (in 3rd Octet)",
                "Valid 4 Octet Sequence",
                "Invalid 4 Octet Sequence (in 2nd Octet)",
                "Invalid 4 Octet Sequence (in 3rd Octet)",
                "Invalid 4 Octet Sequence (in 4th Octet)",
                "Valid 5 Octet Sequence (but not Unicode!)",
                "Valid 6 Octet Sequence (but not Unicode!)",
            ],
        }
    )

    result = df.select(col("bytes").try_decode("utf-8").alias("decoded"), col("note"))

    expected = {
        "decoded": [
            None,
            TEXT.decode("utf-8"),
            str(valid_ascii, "utf-8"),
            str(valid_2_octet, "utf-8"),
            None,  # Invalid 2 Octet Sequence
            None,  # Invalid Sequence Identifier
            str(valid_3_octet, "utf-8"),
            None,  # Invalid 3 Octet Sequence (in 2nd Octet)
            None,  # Invalid 3 Octet Sequence (in 3rd Octet)
            str(valid_4_octet, "utf-8"),
            None,  # Invalid 4 Octet Sequence (in 2nd Octet)
            None,  # Invalid 4 Octet Sequence (in 3rd Octet)
            None,  # Invalid 4 Octet Sequence (in 4th Octet)
            None,  # Valid 5 Octet Sequence (but not Unicode!)
            None,  # Valid 6 Octet Sequence (but not Unicode!)
        ],
        "note": df.to_pydict()["note"],
    }

    assert result.to_pydict() == expected


@pytest.mark.skip("sanity perf checking")
def test_try_decode_utf8_perf():
    from daft import DataType as dt

    @daft.udf(return_dtype=dt.string())
    def try_decode_utf8_udf(binary_series):
        strings = []
        for binary in binary_series:
            try:
                string = binary.decode("utf-8")
                strings.append(string)
            except UnicodeDecodeError:
                strings.append(None)
        return strings

    # Create a large dataset with valid and invalid UTF-8 strings
    n = 1_000_000
    valid_data = [TEXT] * (n // 2)
    invalid_data = [b"\xc3\x28"] * (n // 2)  # Invalid UTF-8 sequence
    mixed_data = valid_data + invalid_data

    df = daft.from_pydict({"bytes": mixed_data})

    import statistics
    import time

    # run tests multiple times for statistical significance
    udf_times = []
    native_times = []
    iterations = 11  # first one is warm-up

    for i in range(iterations):
        # test udf
        start = time.time()
        _ = df.select(try_decode_utf8_udf(col("bytes")).alias("decoded")).collect()
        elapsed = time.time() - start
        if i > 0:  # Skip the first iteration (warm-up)
            udf_times.append(elapsed)
        # test try_decode('utf-8')
        start = time.time()
        _ = df.select(col("bytes").try_decode("utf-8").alias("decoded")).collect()
        elapsed = time.time() - start
        if i > 0:  # Skip the first iteration (warm-up)
            native_times.append(elapsed)

    udf_time = sum(udf_times) / len(udf_times)
    native_time = sum(native_times) / len(native_times)

    # Statistical summary
    udf_stats = {
        "mean": udf_time,
        "median": statistics.median(udf_times),
        "min": min(udf_times),
        "max": max(udf_times),
        "stdev": statistics.stdev(udf_times) if len(udf_times) > 1 else 0,
    }

    native_stats = {
        "mean": native_time,
        "median": statistics.median(native_times),
        "min": min(native_times),
        "max": max(native_times),
        "stdev": statistics.stdev(native_times) if len(native_times) > 1 else 0,
    }

    print(f"Native try_decode stats (seconds): {native_stats}")
    print(f"UDF try_decode stats (seconds): {udf_stats}")
    print(f"Average speedup: {udf_time / native_time:.2f}x")


###
# try_ tests
###


def test_try_encode():
    pass


def test_try_decode():
    import gzip
    import zlib

    df = daft.from_pydict(
        {
            "bytes": [
                None,
                zlib.compress(TEXT),
                gzip.compress(TEXT),
                TEXT,
            ]
        }
    )
    actual = df.select(
        col("bytes").try_decode("zlib").alias("try_zlib"),
        col("bytes").try_decode("gzip").alias("try_gzip"),
    )
    expect = {
        "try_zlib": [None, TEXT, None, None],
        "try_gzip": [None, None, TEXT, None],
    }
    assert actual.to_pydict() == expect



from __future__ import annotations

from daft.datatype import DataType
from daft.recordbatch.micropartition import MicroPartition
from daft.recordbatch.recordbatch import RecordBatch


def test_monotonically_increasing_id() -> None:
    table = MicroPartition.from_pydict({"a": [1, 2, 3, 4, 5]})
    table = table.add_monotonically_increasing_id(0, "id")

    assert len(table) == 5
    assert set(table.column_names()) == {"id", "a"}
    assert set(table.schema().column_names()) == {"id", "a"}
    assert table.schema()["id"].dtype == DataType.uint64()
    assert table.to_pydict() == {"id": [0, 1, 2, 3, 4], "a": [1, 2, 3, 4, 5]}


def test_monotonically_increasing_id_empty_table() -> None:
    table = MicroPartition.from_pydict({"a": []})
    table = table.add_monotonically_increasing_id(0, "id")

    assert len(table) == 0
    assert set(table.column_names()) == {"id", "a"}
    assert set(table.schema().column_names()) == {"id", "a"}
    assert table.schema()["id"].dtype == DataType.uint64()
    assert table.to_pydict() == {"id": [], "a": []}


def test_monotonically_increasing_id_multiple_tables_in_micropartition() -> None:
    table1 = RecordBatch.from_pydict({"a": [1, 2, 3]})
    table2 = RecordBatch.from_pydict({"a": [4, 5, 6]})
    table3 = RecordBatch.from_pydict({"a": [7, 8, 9]})

    table = MicroPartition._from_record_batches([table1, table2, table3])
    table = table.add_monotonically_increasing_id(0, "id")

    assert len(table) == 9
    assert set(table.column_names()) == {"id", "a"}
    assert set(table.schema().column_names()) == {"id", "a"}
    assert table.schema()["id"].dtype == DataType.uint64()
    assert table.to_pydict() == {"id": [0, 1, 2, 3, 4, 5, 6, 7, 8], "a": [1, 2, 3, 4, 5, 6, 7, 8, 9]}



"""Run TPC-H benchmarks with Ray runner on local Parquet data and upload results to Google sheets."""

from __future__ import annotations

import asyncio
import os
import sys
import time
import urllib.request

from ray.job_submission import JobDetails, JobStatus, JobSubmissionClient

import daft
from tools.ci_bench_utils import get_run_metadata, upload_to_google_sheets


async def tail_logs(client: JobSubmissionClient, submission_id: str) -> JobDetails:
    async for lines in client.tail_job_logs(submission_id):
        print(lines, end="")

    return client.get_job_info(submission_id)


SF_TO_S3_PATH = {
    100: "s3://eventual-dev-benchmarking-fixtures/uncompressed/tpch-dbgen/100_0/32/parquet/",
    1000: "s3://eventual-dev-benchmarking-fixtures/uncompressed/tpch-dbgen/1000_0/512/parquet/",
}


def run_benchmark():
    results = {}

    scale_factor = int(os.getenv("TPCH_SCALE_FACTOR"))
    if scale_factor not in SF_TO_S3_PATH:
        raise ValueError(
            f"TPC-H scale factor {scale_factor} not supported, expected one of the following: {SF_TO_S3_PATH.keys()}"
        )
    parquet_path = SF_TO_S3_PATH[scale_factor]

    client = JobSubmissionClient(address="http://localhost:8265")

    # Create the directory if it doesn't exist
    os.makedirs("traces", exist_ok=True)

    for q in range(1, 23):
        print(f"Running TPC-H Q{q}... ", end="", flush=True)

        start = time.perf_counter()

        submission_id = client.submit_job(
            entrypoint=f"DAFT_RUNNER=ray DAFT_PROGRESS_BAR=0 python answers_sql.py {parquet_path} {q}",
            runtime_env={"working_dir": "./benchmarking/tpch"},
        )

        job_details = asyncio.run(tail_logs(client, submission_id))

        end = time.perf_counter()

        # Download the trace file
        trace_url = f"http://localhost:8265/api/v0/tasks/timeline?download=1&job_id={job_details.job_id}"
        urllib.request.urlretrieve(trace_url, f"traces/q{q}-trace.json")

        if job_details.status != JobStatus.SUCCEEDED:
            print(f"\nRay job did not succeed, received job status: {job_details.status}\nJob details: {job_details}")
            sys.exit(1)

        results[q] = end - start

        print(f"done in {results[q]:.2f}s")
        print(f"Job details: {job_details}")

    return results


def main():
    daft.context.set_runner_native()

    metadata = get_run_metadata()
    scale_factor = int(os.getenv("TPCH_SCALE_FACTOR"))
    num_workers = int(os.getenv("RAY_NUM_WORKERS"))

    print("Starting warmup run...")
    run_benchmark()

    print("Warmup done. Running benchmark and collecting results...")
    results = run_benchmark()

    data_dict = {**metadata, "scale factor": scale_factor, "num workers": num_workers, **results}

    print("Results:")
    print(data_dict)

    upload_to_google_sheets("Distributed TPC-H", list(data_dict.values()))


if __name__ == "__main__":
    main()



from __future__ import annotations

import base64

import pytest
import tiktoken

import daft
from daft import col

TOKEN_FILE = "tests/assets/tokens/tokens_5k.tiktoken"
HTTP_TOKEN_FILE = (
    "https://daft-public-data.s3.us-west-2.amazonaws.com/test_fixtures/tiktoken_testing/p50k_base.tiktoken"
)
AWS_TOKEN_FILE = "s3://daft-public-data/test_fixtures/tiktoken_testing/p50k_base.tiktoken"
P50K_REGEX = "'(?:[sdmt]|ll|ve|re)| ?\\p{L}+| ?\\p{N}+| ?[^\\s\\p{L}\\p{N}]+|\\s+(?!\\S)|\\s+"


def get_openai_enc():
    # from https://github.com/openai/tiktoken/blob/main/tiktoken/load.py
    with open(TOKEN_FILE, "rb") as f:
        file_bytes = f.read()
    enc_dict = {
        base64.b64decode(token): int(rank) for token, rank in (line.split() for line in file_bytes.splitlines() if line)
    }
    return tiktoken.Encoding("test_encoding", pat_str=P50K_REGEX, mergeable_ranks=enc_dict, special_tokens={})


openai_enc = get_openai_enc()


def test_file_token_encode():
    test_data = ["hello custom tokenizer!", "hopefully this works", "", "wow!"]
    df = daft.from_pydict({"a": test_data})
    res = df.select(col("a").str.tokenize_encode(TOKEN_FILE, pattern=P50K_REGEX)).to_pydict()["a"]

    openai_res = openai_enc.encode_batch(test_data)
    assert res == openai_res


def test_file_token_decode():
    test_data = ["hello custom tokenizer!", "hopefully this works", "", "wow!"]
    token_data = openai_enc.encode_batch(test_data)
    df = daft.from_pydict({"a": token_data})
    res = df.select(col("a").str.tokenize_decode(TOKEN_FILE, pattern=P50K_REGEX)).to_pydict()["a"]
    assert res == test_data


@pytest.mark.integration()
def test_file_http_download():
    test_data = ["hello custom tokenizer!", "hopefully this works", "", "wow!"]
    df = daft.from_pydict({"a": test_data})
    res = df.select(col("a").str.tokenize_encode(HTTP_TOKEN_FILE, pattern=P50K_REGEX)).to_pydict()["a"]

    http_openai_enc = tiktoken.get_encoding("p50k_base")
    openai_res = http_openai_enc.encode_batch(test_data)
    assert res == openai_res

    df = daft.from_pydict({"b": openai_res})
    res = df.select(col("b").str.tokenize_decode(HTTP_TOKEN_FILE, pattern=P50K_REGEX)).to_pydict()["b"]
    assert res == test_data


@pytest.mark.integration()
def test_file_aws_download():
    test_data = ["hello custom tokenizer!", "hopefully this works", "", "wow!"]
    df = daft.from_pydict({"a": test_data})
    res = df.select(col("a").str.tokenize_encode(AWS_TOKEN_FILE, pattern=P50K_REGEX)).to_pydict()["a"]

    aws_openai_enc = tiktoken.get_encoding("p50k_base")
    openai_res = aws_openai_enc.encode_batch(test_data)
    assert res == openai_res

    df = daft.from_pydict({"b": openai_res})
    res = df.select(col("b").str.tokenize_decode(AWS_TOKEN_FILE, pattern=P50K_REGEX)).to_pydict()["b"]
    assert res == test_data


@pytest.mark.integration()
def test_file_aws_download_ioconfig():
    io_config = daft.io.IOConfig(
        s3=daft.io.S3Config(
            region_name="us-west-2",
            anonymous=True,
        )
    )
    test_data = ["hello custom tokenizer!", "hopefully this works", "", "wow!"]
    df = daft.from_pydict({"a": test_data})
    res = df.select(col("a").str.tokenize_encode(AWS_TOKEN_FILE, io_config=io_config, pattern=P50K_REGEX)).to_pydict()[
        "a"
    ]

    aws_openai_enc = tiktoken.get_encoding("p50k_base")
    openai_res = aws_openai_enc.encode_batch(test_data)
    assert res == openai_res

    df = daft.from_pydict({"b": openai_res})
    res = df.select(col("b").str.tokenize_decode(AWS_TOKEN_FILE, io_config=io_config, pattern=P50K_REGEX)).to_pydict()[
        "b"
    ]
    assert res == test_data



from __future__ import annotations

import pytest

import daft
import daft.exceptions
from daft import col


def test_to_struct_empty_structs():
    df = daft.from_pydict({"a": [1, 2, 3]})

    with pytest.raises(daft.exceptions.DaftCoreException, match="Cannot call struct with no inputs"):
        df.select(daft.struct()).collect()


# there was a bug with pushdowns onto to_struct previously
def test_to_struct_pushdown():
    df = daft.from_pydict(
        {
            "a": [1, 2, 3, 4, None, 6, None],
            "b": ["a", "b", "c", "", "e", None, None],
        }
    )
    df = df.select(daft.struct(col("a"), col("b")))
    df = df.select(col("struct").struct.get("a"))
    assert df.to_pydict() == {"a": [1, 2, 3, 4, None, 6, None]}



from __future__ import annotations

import pyarrow as pa
import pyarrow.parquet as papq
import pytest

import daft


@pytest.fixture(scope="function")
def parquet_files(tmpdir):
    """Writes 1 Parquet file with 10 rowgroups, each of 100 bytes in size."""
    tbl = pa.table({"data": ["aaa"] * 100})
    path = tmpdir / "file.pq"
    papq.write_table(tbl, str(path), row_group_size=10, use_dictionary=False)

    return tmpdir


def test_split_parquet_read(parquet_files):
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=1,
        scan_tasks_max_size_bytes=10,
    ):
        df = daft.read_parquet(str(parquet_files))
        assert df.num_partitions() == 10, "Should have 10 partitions since we will split the file"
        assert df.to_pydict() == {"data": ["aaa"] * 100}


@pytest.mark.skip(reason="Not implemented yet")
def test_split_parquet_read_some_splits(tmpdir):
    with daft.execution_config_ctx(scantask_splitting_level=2):
        # Write a mix of 20 large and 20 small files
        # Small ones should not be split, large ones should be split into 10 rowgroups each
        # This gives us a total of 200 + 20 scantasks

        # Write 20 large files into tmpdir
        large_file_paths = []
        for i in range(20):
            tbl = pa.table({"data": [str(f"large{i}") for i in range(100)]})
            path = tmpdir / f"file.{i}.large.pq"
            papq.write_table(tbl, str(path), row_group_size=10, use_dictionary=False)
            large_file_paths.append(str(path))

        # Write 20 small files into tmpdir
        small_file_paths = []
        for i in range(20):
            tbl = pa.table({"data": ["small"]})
            path = tmpdir / f"file.{i}.small.pq"
            papq.write_table(tbl, str(path), row_group_size=1, use_dictionary=False)
            small_file_paths.append(str(path))

        # Test [large_paths, ..., small_paths, ...]
        with daft.execution_config_ctx(
            scan_tasks_min_size_bytes=20,
            scan_tasks_max_size_bytes=100,
        ):
            df = daft.read_parquet(large_file_paths + small_file_paths)
            assert (
                df.num_partitions() == 220
            ), "Should have 220 partitions since we will split all large files (20 * 10 rowgroups) but keep small files unsplit"
            assert df.to_pydict() == {"data": [str(f"large{i}") for i in range(100)] * 20 + ["small"] * 20}

        # Test interleaved [large_path, small_path, large_path, small_path, ...]
        with daft.execution_config_ctx(
            scan_tasks_min_size_bytes=20,
            scan_tasks_max_size_bytes=100,
        ):
            interleaved_paths = [path for pair in zip(large_file_paths, small_file_paths) for path in pair]
            df = daft.read_parquet(interleaved_paths)
            assert (
                df.num_partitions() == 220
            ), "Should have 220 partitions since we will split all large files (20 * 10 rowgroups) but keep small files unsplit"
            assert df.to_pydict() == {"data": ([str(f"large{i}") for i in range(100)] + ["small"]) * 20}

        # Test [small_paths, ..., large_paths]
        with daft.execution_config_ctx(
            scan_tasks_min_size_bytes=20,
            scan_tasks_max_size_bytes=100,
        ):
            df = daft.read_parquet(small_file_paths + large_file_paths)
            assert (
                df.num_partitions() == 220
            ), "Should have 220 partitions since we will split all large files (20 * 10 rowgroups) but keep small files unsplit"
            assert df.to_pydict() == {"data": ["small"] * 20 + [str(f"large{i}") for i in range(100)] * 20}



from __future__ import annotations


def test_create_simple_df(spark_session):
    # Create simple DataFrame with single column
    data = [(1,), (2,), (3,)]
    df = spark_session.createDataFrame(data, ["id"])

    # Convert to pandas and verify
    df_pandas = df.toPandas()
    assert len(df_pandas) == 3, "DataFrame should have 3 rows"
    assert list(df_pandas["id"]) == [1, 2, 3], "DataFrame should contain expected values"


def test_create_float_df(spark_session):
    # Create DataFrame with float column
    float_data = [(1.1,), (2.2,), (3.3,)]
    df_float = spark_session.createDataFrame(float_data, ["value"])
    df_float_pandas = df_float.toPandas()
    assert len(df_float_pandas) == 3, "Float DataFrame should have 3 rows"
    assert list(df_float_pandas["value"]) == [1.1, 2.2, 3.3], "Float DataFrame should contain expected values"


def test_create_two_column_df(spark_session):
    # Create DataFrame with two numeric columns
    two_col_data = [(1, 10), (2, 20), (3, 30)]
    df_two = spark_session.createDataFrame(two_col_data, ["num1", "num2"])
    df_two_pandas = df_two.toPandas()
    assert len(df_two_pandas) == 3, "Two-column DataFrame should have 3 rows"
    assert list(df_two_pandas["num1"]) == [1, 2, 3], "First number column should contain expected values"
    assert list(df_two_pandas["num2"]) == [10, 20, 30], "Second number column should contain expected values"


def test_create_boolean_df(spark_session):
    boolean_data = [(True,), (False,), (True,)]
    df_boolean = spark_session.createDataFrame(boolean_data, ["value"])
    df_boolean_pandas = df_boolean.toPandas()
    assert len(df_boolean_pandas) == 3, "Boolean DataFrame should have 3 rows"
    assert list(df_boolean_pandas["value"]) == [True, False, True], "Boolean DataFrame should contain expected values"


def test_create_string_df(spark_session):
    # Create DataFrame with string column
    string_data = [("hello",), ("world",), ("test",)]
    df_string = spark_session.createDataFrame(string_data, ["text"])
    df_string_pandas = df_string.toPandas()
    assert len(df_string_pandas) == 3, "String DataFrame should have 3 rows"
    assert list(df_string_pandas["text"]) == [
        "hello",
        "world",
        "test",
    ], "String DataFrame should contain expected values"


def test_create_mixed_type_df(spark_session):
    # Create DataFrame with mixed types
    mixed_data = [(1, "one", True), (2, "two", False), (3, "three", True)]
    df_mixed = spark_session.createDataFrame(mixed_data, ["id", "name", "active"])
    df_mixed_pandas = df_mixed.toPandas()
    assert len(df_mixed_pandas) == 3, "Mixed-type DataFrame should have 3 rows"
    assert list(df_mixed_pandas["id"]) == [1, 2, 3], "ID column should contain expected values"
    assert list(df_mixed_pandas["name"]) == ["one", "two", "three"], "Name column should contain expected values"
    assert list(df_mixed_pandas["active"]) == [True, False, True], "Active column should contain expected values"


def test_create_null_df(spark_session):
    # Create DataFrame with null values
    null_data = [(1, None), (None, "test"), (3, "data")]
    df_null = spark_session.createDataFrame(null_data, ["id", "value"])
    df_null_pandas = df_null.toPandas()
    assert len(df_null_pandas) == 3, "Null-containing DataFrame should have 3 rows"
    assert df_null_pandas["id"].isna().sum() == 1, "ID column should have one null value"
    assert df_null_pandas["value"].isna().sum() == 1, "Value column should have one null value"


def test_create_decimal_df(spark_session):
    # Create DataFrame with decimal/double precision numbers
    decimal_data = [(1.23456789,), (9.87654321,), (5.55555555,)]
    df_decimal = spark_session.createDataFrame(decimal_data, ["amount"])
    df_decimal_pandas = df_decimal.toPandas()
    assert len(df_decimal_pandas) == 3, "Decimal DataFrame should have 3 rows"
    assert abs(df_decimal_pandas["amount"][0] - 1.23456789) < 1e-8, "Decimal values should maintain precision"


def test_create_special_chars_df(spark_session):
    # Create DataFrame with empty strings and special characters
    special_data = [("",), (" ",), ("!@#$%^&*",)]
    df_special = spark_session.createDataFrame(special_data, ["special"])
    df_special_pandas = df_special.toPandas()
    assert len(df_special_pandas) == 3, "Special character DataFrame should have 3 rows"
    assert list(df_special_pandas["special"]) == [
        "",
        " ",
        "!@#$%^&*",
    ], "Special character DataFrame should contain expected values"



# ruff: noqa: I002
# isort: dont-add-import: from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Union

from daft.api_annotations import PublicAPI

if TYPE_CHECKING:
    import dask
    import numpy as np
    import pandas as pd
    import pyarrow as pa
    from ray.data.dataset import Dataset as RayDataset

    from daft.dataframe import DataFrame


InputListType = Union[list[Any], "np.ndarray[Any, Any]", "pa.Array", "pa.ChunkedArray"]


@PublicAPI
def from_pylist(data: list[dict[str, Any]]) -> "DataFrame":
    """Creates a DataFrame from a list of dictionaries.

    Args:
        data: List of dictionaries, where each key is a column name.

    Returns:
        DataFrame: DataFrame created from list of dictionaries.

    Examples:
        >>> import daft
        >>> df = daft.from_pylist([{"foo": 1}, {"foo": 2}])
        >>> df.show()
        ╭───────╮
        │ foo   │
        │ ---   │
        │ Int64 │
        ╞═══════╡
        │ 1     │
        ├╌╌╌╌╌╌╌┤
        │ 2     │
        ╰───────╯
        <BLANKLINE>
        (Showing first 2 of 2 rows)
    """
    from daft import DataFrame

    return DataFrame._from_pylist(data)


@PublicAPI
def from_pydict(data: dict[str, InputListType]) -> "DataFrame":
    """Creates a DataFrame from a Python dictionary.

    Args:
        data: Key -> Sequence[item] of data. Each Key is created as a column, and must have a value that is
            a Python list, Numpy array or PyArrow array. Values must be equal in length across all keys.

    Returns:
        DataFrame: DataFrame created from dictionary of columns

    Examples:
        >>> import daft
        >>> df = daft.from_pydict({"foo": [1, 2]})
        >>> df.show()
        ╭───────╮
        │ foo   │
        │ ---   │
        │ Int64 │
        ╞═══════╡
        │ 1     │
        ├╌╌╌╌╌╌╌┤
        │ 2     │
        ╰───────╯
        <BLANKLINE>
        (Showing first 2 of 2 rows)
    """
    from daft import DataFrame

    return DataFrame._from_pydict(data)


@PublicAPI
def from_arrow(data: Union["pa.Table", list["pa.Table"], Iterable["pa.Table"]]) -> "DataFrame":
    """Creates a DataFrame from a pyarrow Table.

    Args:
        data: pyarrow Table(s) that we wish to convert into a Daft DataFrame.

    Returns:
        DataFrame: DataFrame created from the provided pyarrow Table.

    Examples:
        >>> import pyarrow as pa
        >>> import daft
        >>> t = pa.table({"a": [1, 2, 3], "b": ["foo", "bar", "baz"]})
        >>> df = daft.from_arrow(t)
        >>> df.show()
        ╭───────┬──────╮
        │ a     ┆ b    │
        │ ---   ┆ ---  │
        │ Int64 ┆ Utf8 │
        ╞═══════╪══════╡
        │ 1     ┆ foo  │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌┤
        │ 2     ┆ bar  │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌┤
        │ 3     ┆ baz  │
        ╰───────┴──────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    from daft import DataFrame

    return DataFrame._from_arrow(data)


@PublicAPI
def from_pandas(data: Union["pd.DataFrame", list["pd.DataFrame"]]) -> "DataFrame":
    """Creates a Daft DataFrame from a pandas DataFrame.

    Args:
        data: pandas DataFrame(s) that we wish to convert into a Daft DataFrame.

    Returns:
        DataFrame: Daft DataFrame created from the provided pandas DataFrame.

    Examples:
        >>> import pandas as pd
        >>> import daft
        >>> pd_df = pd.DataFrame({"a": [1, 2, 3], "b": ["foo", "bar", "baz"]})
        >>> df = daft.from_pandas(pd_df)
        >>> df.show()
        ╭───────┬──────╮
        │ a     ┆ b    │
        │ ---   ┆ ---  │
        │ Int64 ┆ Utf8 │
        ╞═══════╪══════╡
        │ 1     ┆ foo  │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌┤
        │ 2     ┆ bar  │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌┤
        │ 3     ┆ baz  │
        ╰───────┴──────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    from daft import DataFrame

    return DataFrame._from_pandas(data)


@PublicAPI
def from_ray_dataset(ds: "RayDataset") -> "DataFrame":
    """Creates a DataFrame from a Ray Dataset.

    Args:
        ds: The Ray Dataset to create a Daft DataFrame from.

    Returns:
        DataFrame: Daft DataFrame created from the provided Ray dataset.

    Note:
        This function can only work if Daft is running using the RayRunner

    """
    from daft import DataFrame

    return DataFrame._from_ray_dataset(ds)


@PublicAPI
def from_dask_dataframe(ddf: "dask.DataFrame") -> "DataFrame":
    """Creates a Daft DataFrame from a Dask DataFrame.

    The provided Dask DataFrame must have been created using [Dask-on-Ray](https://docs.ray.io/en/latest/ray-more-libs/dask-on-ray.html).

    Args:
        ddf: The Dask DataFrame to create a Daft DataFrame from.

    Note:
        This function can only work if Daft is running using the RayRunner

    """
    from daft import DataFrame

    return DataFrame._from_dask_dataframe(ddf)



from __future__ import annotations

import numpy as np
import pytest

import daft
from daft import DataFrame


@pytest.fixture(scope="module")
def gen_aranged_df(num_samples=1_000_000) -> DataFrame:
    return daft.from_pydict(
        {
            "i": ((np.arange(num_samples, dtype=np.int64) * 9582398353) % 100),
            "j": ((np.arange(num_samples, dtype=np.int64) * 847892347987) % 100),
        }
    ).collect()


@pytest.mark.benchmark(group="arithmetic")
def test_integer_multiplications(gen_aranged_df, benchmark) -> None:
    """Integer multiplications between 1_000_000 values.

    Adapted from: https://github.com/duckdb/duckdb/blob/master/benchmark/micro/arithmetic/multiplications.benchmark
    """

    def bench_sum() -> DataFrame:
        return (
            gen_aranged_df.with_column(
                "k",
                (gen_aranged_df["i"] * gen_aranged_df["j"])
                + (gen_aranged_df["i"] * gen_aranged_df["j"])
                + (gen_aranged_df["i"] * gen_aranged_df["j"])
                + (gen_aranged_df["i"] * gen_aranged_df["j"]),
            )
            .min("k")
            .collect()
        )

    result = benchmark(bench_sum)
    assert result.to_pydict()["k"] == [0]



# ruff: noqa: I002
# isort: dont-add-import: from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING, Optional

from daft import context
from daft.api_annotations import PublicAPI
from daft.daft import IOConfig, PyPartitionField, PyPushdowns, PyRecordBatch, ScanOperatorHandle, ScanTask
from daft.dataframe import DataFrame
from daft.io.object_store_options import io_config_to_storage_options
from daft.io.scan import ScanOperator
from daft.logical.builder import LogicalPlanBuilder
from daft.logical.schema import Schema
from daft.recordbatch import RecordBatch

if TYPE_CHECKING:
    import lance


def _lancedb_table_factory_function(
    ds: "lance.LanceDataset", fragment_id: int, required_columns: Optional[list[str]]
) -> Iterator[PyRecordBatch]:
    fragment = ds.get_fragment(fragment_id)
    assert fragment is not None, RuntimeError(f"Unable to find lance fragment {fragment_id}")
    return (
        RecordBatch.from_arrow_record_batches([rb], rb.schema)._recordbatch
        for rb in fragment.to_batches(columns=required_columns)
    )


@PublicAPI
def read_lance(url: str, io_config: Optional[IOConfig] = None) -> DataFrame:
    """Create a DataFrame from a LanceDB table.

    Args:
        url: URL to the LanceDB table (supports remote URLs to object stores such as `s3://` or `gs://`)
        io_config: A custom IOConfig to use when accessing LanceDB data. Defaults to None.

    Returns:
        DataFrame: a DataFrame with the schema converted from the specified LanceDB table

    Note:
        This function requires the use of [LanceDB](https://lancedb.github.io/lancedb/), which is the Python library for the LanceDB project.
        To ensure that this is installed with Daft, you may install: `pip install daft[lance]`

    Examples:
        Read a local LanceDB table:
        >>> df = daft.read_lance("s3://my-lancedb-bucket/data/")
        >>> df.show()

        Read a LanceDB table from a public S3 bucket:
        >>> from daft.io import S3Config
        >>> s3_config = S3Config(region="us-west-2", anonymous=True)
        >>> df = daft.read_lance("s3://daft-public-data/lance/words-test-dataset", io_config=s3_config)
        >>> df.show()
    """
    try:
        import lance
    except ImportError as e:
        raise ImportError(
            "Unable to import the `lance` package, please ensure that Daft is installed with the lance extra dependency: `pip install daft[lance]`"
        ) from e

    io_config = context.get_context().daft_planning_config.default_io_config if io_config is None else io_config
    storage_options = io_config_to_storage_options(io_config, url)

    ds = lance.dataset(url, storage_options=storage_options)
    iceberg_operator = LanceDBScanOperator(ds)

    handle = ScanOperatorHandle.from_python_scan_operator(iceberg_operator)
    builder = LogicalPlanBuilder.from_tabular_scan(scan_operator=handle)
    return DataFrame(builder)


class LanceDBScanOperator(ScanOperator):
    def __init__(self, ds: "lance.LanceDataset"):
        self._ds = ds

    def name(self) -> str:
        return "LanceDBScanOperator"

    def display_name(self) -> str:
        return f"LanceDBScanOperator({self._ds.uri})"

    def schema(self) -> Schema:
        return Schema.from_pyarrow_schema(self._ds.schema)

    def partitioning_keys(self) -> list[PyPartitionField]:
        return []

    def can_absorb_filter(self) -> bool:
        return False

    def can_absorb_limit(self) -> bool:
        return False

    def can_absorb_select(self) -> bool:
        return False

    def multiline_display(self) -> list[str]:
        return [
            self.display_name(),
            f"Schema = {self.schema()}",
        ]

    def to_scan_tasks(self, pushdowns: PyPushdowns) -> Iterator[ScanTask]:
        required_columns: Optional[list[str]]
        if pushdowns.columns is None:
            required_columns = None
        else:
            filter_required_column_names = pushdowns.filter_required_column_names()
            required_columns = (
                pushdowns.columns
                if filter_required_column_names is None
                else pushdowns.columns + filter_required_column_names
            )

        # TODO: figure out how to translate Pushdowns into LanceDB filters
        filters = None
        fragments = self._ds.get_fragments(filter=filters)
        for i, fragment in enumerate(fragments):
            # TODO: figure out how if we can get this metadata from LanceDB fragments cheaply
            size_bytes = None
            stats = None

            # NOTE: `fragment.count_rows()` should result in 1 IO call for the data file
            # (1 fragment = 1 data file) and 1 more IO call for the deletion file (if present).
            # This could potentially be expensive to perform serially if there are thousands of files.
            # Given that num_rows isn't leveraged for much at the moment, and without statistics
            # we will probably end up materializing the data anyways for any operations, we leave this
            # as None.
            num_rows = None

            yield ScanTask.python_factory_func_scan_task(
                module=_lancedb_table_factory_function.__module__,
                func_name=_lancedb_table_factory_function.__name__,
                func_args=(self._ds, fragment.fragment_id, required_columns),
                schema=self.schema()._schema,
                num_rows=num_rows,
                size_bytes=size_bytes,
                pushdowns=pushdowns,
                stats=stats,
            )
